from collections.abc import Collection
from unittest.mock import MagicMock

import pytest

import briefcase.commands.new
from briefcase.bootstraps import (
    BaseGuiBootstrap,
    ConsoleBootstrap,
    EmptyBootstrap,
    PygameGuiBootstrap,
    PySide6GuiBootstrap,
    TogaGuiBootstrap,
)


@pytest.fixture
def mock_builtin_bootstraps():
    return {
        "None": EmptyBootstrap,
        "Toga": TogaGuiBootstrap,
        "Console": ConsoleBootstrap,
        "PySide6": PySide6GuiBootstrap,
        "Pygame": PygameGuiBootstrap,
    }


def test_question_sequence_bootstrap_context(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """The context passed to the bootstrap is correct."""

    passed_context = {}

    class GuiBootstrap:
        fields: Collection[str] = []

        def __init__(self, console, context):
            nonlocal passed_context
            passed_context = context.copy()

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value={
                **mock_builtin_bootstraps,
                "Custom GUI": GuiBootstrap,
            },
        ),
    )

    new_command.console.input_enabled = False

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={
            "bootstrap": "Custom GUI",
        },
    )

    assert isinstance(bootstrap, GuiBootstrap)
    assert passed_context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_toga(new_command):
    """The Toga bootstrap can be selected."""

    new_command.console.values = [
        "1",  # Toga GUI toolkit
    ]

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, TogaGuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_console(new_command):
    """A console bootstrap can be constructed."""

    new_command.console.values = [
        "4",  # Console app
    ]

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, ConsoleBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_pyside6(new_command):
    """A Pyside6 bootstrap can be created."""

    new_command.console.values = [
        "2",  # PySide6 GUI toolkit
    ]

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, PySide6GuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_pygame(new_command):
    """A Pygame bootstrap can be constructed."""

    new_command.console.values = [
        "3",  # Pygame GUI toolkit
    ]

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, PygameGuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_none(new_command):
    """If no bootstrap is selected, the empty bootstrap is used."""

    # Prime answers for all the questions.
    new_command.console.values = [
        "5",  # None
    ]

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, BaseGuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_with_overrides(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """The answer to the bootstrap question can be overridden."""

    # Prime answers for none of the questions.
    new_command.console.values = []

    class GuiBootstrap:
        fields: Collection[str] = []

        def __init__(self, console, context):
            self.context = context.copy()

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value={
                **mock_builtin_bootstraps,
                "Custom GUI": GuiBootstrap,
            },
        ),
    )

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={
            "bootstrap": "Custom GUI",
        },
    )

    assert isinstance(bootstrap, GuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_with_bad_bootstrap_override(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """A bad override for the bootstrap uses user input instead."""

    # Prime a bad answer for the bootstrap question
    new_command.console.values = [
        "6",  # None
    ]

    class GuiBootstrap:
        # if this custom bootstrap is chosen, the lack of
        # requires() would cause an error
        fields: Collection[str] = ["requires"]

        def __init__(self, console, context):
            pass

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value={
                **mock_builtin_bootstraps,
                "Custom GUI": GuiBootstrap,
            },
        ),
    )

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={
            "bootstrap": "BAD i don't exist GUI",
        },
    )

    assert isinstance(bootstrap, BaseGuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }


def test_question_sequence_with_no_user_input(new_command):
    """If no user input is provided, all user inputs are taken as default."""

    new_command.console.input_enabled = False

    bootstrap = new_command.create_bootstrap(
        context={
            "app_name": "myapplication",
            "author": "Grace Hopper",
        },
        project_overrides={},
    )

    assert isinstance(bootstrap, TogaGuiBootstrap)
    assert bootstrap.context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
    }



import sys
from unittest.mock import MagicMock

import pytest

import briefcase.platforms.linux.appimage
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.linuxdeploy import LinuxDeploy
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.appimage import (
    LinuxAppImageBuildCommand,
    LinuxAppImageCreateCommand,
)


@pytest.fixture
def create_command(dummy_console, tmp_path):
    command = LinuxAppImageCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    # Mock verified linuxdeploy
    command.tools.linuxdeploy = LinuxDeploy(tools=command.tools)
    return command


def test_binary_path(create_command, first_app_config, tmp_path):
    # Force the architecture to x86_64 for test purposes.
    create_command.tools.host_arch = "x86_64"
    binary_path = create_command.binary_path(first_app_config)

    assert (
        binary_path
        == tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage"
    )


def test_project_path(create_command, first_app_config, tmp_path):
    """The project path is the bundle path."""
    project_path = create_command.project_path(first_app_config)
    bundle_path = create_command.bundle_path(first_app_config)

    expected_path = tmp_path / "base_path/build/first-app/linux/appimage"
    assert expected_path == project_path == bundle_path


def test_distribution_path(create_command, first_app_config, tmp_path):
    # Force the architecture to x86_64 for test purposes.
    create_command.tools.host_arch = "x86_64"
    distribution_path = create_command.distribution_path(first_app_config)

    assert (
        distribution_path == tmp_path / "base_path/dist/First_App-0.0.1-x86_64.AppImage"
    )


@pytest.mark.parametrize(
    ("manylinux", "tag"),
    [
        (None, "appimage"),
        ("manylinux1", "manylinux1-appimage"),
        ("manylinux_2_28", "manylinux_2_28-appimage"),
    ],
)
def test_docker_image_tag(create_command, first_app_config, manylinux, tag):
    if manylinux:
        first_app_config.manylinux = manylinux

    image_tag = create_command.docker_image_tag(first_app_config)

    assert image_tag == f"briefcase/com.example.first-app:{tag}"


def test_docker_image_tag_uppercase_name(
    create_command,
    uppercase_app_config,
    tmp_path,
):
    image_tag = create_command.docker_image_tag(uppercase_app_config)

    assert image_tag == "briefcase/com.example.first-app:appimage"


def test_verify_linux_no_docker(create_command, first_app_config, tmp_path):
    """If Docker is disabled on Linux, the app_context is Subprocess."""
    create_command.tools.host_os = "Linux"
    create_command.use_docker = False

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # No error and Subprocess is used.
    assert isinstance(create_command.tools[first_app_config].app_context, Subprocess)
    # Docker is not verified.
    assert not hasattr(create_command.tools, "docker")


def test_verify_linux_docker(create_command, first_app_config, tmp_path, monkeypatch):
    """If Docker is enabled on Linux, the Docker alias is set."""
    create_command.tools.host_os = "Linux"
    create_command.use_docker = True
    create_command.extra_docker_build_args = ["--option-one", "--option-two"]

    # Mock Docker tool verification
    mock__version_compat = MagicMock(spec=Docker._version_compat)
    mock__user_access = MagicMock(spec=Docker._user_access)
    mock__buildx_installed = MagicMock(spec=Docker._buildx_installed)
    mock__is_user_mapping_enabled = MagicMock(spec=Docker._is_user_mapping_enabled)
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_version_compat",
        mock__version_compat,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_user_access",
        mock__user_access,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_buildx_installed",
        mock__buildx_installed,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_is_user_mapping_enabled",
        mock__is_user_mapping_enabled,
    )
    mock_docker_app_context_verify = MagicMock(spec=DockerAppContext.verify)
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.DockerAppContext,
        "verify",
        mock_docker_app_context_verify,
    )

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # Docker and Docker app context are verified
    mock__version_compat.assert_called_with(tools=create_command.tools)
    mock__user_access.assert_called_with(tools=create_command.tools)
    mock__buildx_installed.assert_called_with(tools=create_command.tools)
    mock__is_user_mapping_enabled.assert_called_with(None)
    assert isinstance(create_command.tools.docker, Docker)
    mock_docker_app_context_verify.assert_called_with(
        tools=create_command.tools,
        app=first_app_config,
        image_tag="briefcase/com.example.first-app:appimage",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
        extra_build_args=["--option-one", "--option-two"],
    )


def test_verify_non_linux_docker(
    create_command,
    first_app_config,
    monkeypatch,
    tmp_path,
):
    """If Docker is enabled on non-Linux, the Docker alias is set."""
    create_command.tools.host_os = "Darwin"
    create_command.use_docker = True
    create_command.extra_docker_build_args = ["--option-one", "--option-two"]

    # Mock Docker tool verification
    mock__version_compat = MagicMock(spec=Docker._version_compat)
    mock__user_access = MagicMock(spec=Docker._user_access)
    mock__buildx_installed = MagicMock(spec=Docker._buildx_installed)
    mock__is_user_mapping_enabled = MagicMock(spec=Docker._is_user_mapping_enabled)
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_version_compat",
        mock__version_compat,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_user_access",
        mock__user_access,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_buildx_installed",
        mock__buildx_installed,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.Docker,
        "_is_user_mapping_enabled",
        mock__is_user_mapping_enabled,
    )
    mock_docker_app_context_verify = MagicMock(spec=DockerAppContext.verify)
    monkeypatch.setattr(
        briefcase.platforms.linux.appimage.DockerAppContext,
        "verify",
        mock_docker_app_context_verify,
    )

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # Docker and Docker app context are verified
    mock__version_compat.assert_called_with(tools=create_command.tools)
    mock__user_access.assert_called_with(tools=create_command.tools)
    mock__buildx_installed.assert_called_with(tools=create_command.tools)
    mock__is_user_mapping_enabled.assert_called_with(None)
    assert isinstance(create_command.tools.docker, Docker)
    mock_docker_app_context_verify.assert_called_with(
        tools=create_command.tools,
        app=first_app_config,
        image_tag="briefcase/com.example.first-app:appimage",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
        extra_build_args=["--option-one", "--option-two"],
    )


def test_clone_options(dummy_console, tmp_path):
    """Docker options are cloned."""
    build_command = LinuxAppImageBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    build_command.use_docker = True
    build_command.extra_docker_build_args = ["--option-one", "--option-two"]

    create_command = build_command.create_command

    # Confirm the use_docker option has been cloned
    assert create_command.is_clone
    assert create_command.use_docker
    assert create_command.extra_docker_build_args == ["--option-one", "--option-two"]



import os
import shutil
import subprocess as stdlib_subprocess
import sys
from pathlib import Path

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations import subprocess
from briefcase.integrations.base import Tool, ToolCache
from briefcase.integrations.subprocess import SubprocessArgsT


class VenvContext:
    def __init__(self, tools: ToolCache, venv_path: Path):
        self.tools = tools
        self.venv_path = venv_path
        self.created = False

    @property
    def bin_dir(self) -> Path:
        """Return the path to the virtual environment's binary directory.

        :returns: The ``/bin`` (``\\Scripts`` on Windows) directory of the venv.
        """

        return self.venv_path / ("Scripts" if os.name == "nt" else "bin")

    @property
    def executable(self) -> str:
        """Path to the Python executable in the virtual environment.

        :returns: Absolute filesystem path to the venv's Python executable.
        """
        python = self.bin_dir / ("python.exe" if os.name == "nt" else "python")
        return python

    def exists(self) -> bool:
        """Check whether the virtual environment exists.

        :returns: True if venv exists and contains a pyvenv.cfg file.
        """
        return self.venv_path.exists() and (self.venv_path / "pyvenv.cfg").exists()

    def create(self) -> None:
        """Create the virtual environment.

        :raises: BriefcaseCommandError if venv creation fails.
        """
        try:
            self.venv_path.parent.mkdir(parents=True, exist_ok=True)
            self.tools.subprocess.run(
                [sys.executable, "-m", "venv", self.venv_path],
                check=True,
            )
            self.update_core_tools()
            self.created = True
        except stdlib_subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Failed to create virtual environment at {self.venv_path}"
            ) from e

    def clean(self) -> None:
        """Remove the virtual environment."""
        if self.exists():
            shutil.rmtree(self.venv_path)

    def recreate(self) -> None:
        """Remove and re-create the virtual environment."""
        self.clean()
        self.create()

    def update_core_tools(self) -> None:
        """Upgrade core Python packaging tooling in the venv."""
        try:
            self.run([self.executable, "-m", "pip", "install", "-U", "pip"], check=True)
        except Exception as e:
            raise BriefcaseCommandError(
                f"Failed to update core tooling for {self.venv_path}"
            ) from e

    def _rewrite_head(self, args: SubprocessArgsT) -> SubprocessArgsT:
        """Rewrite the first argument to ensure it points to the venv's Python
        executable.

        :param args: Original subprocess arguments.
        :returns: Updated subprocess arguments.
        """
        if not args:
            return args
        head = os.fspath(args[0])
        if os.path.normcase(head) == os.path.normcase(sys.executable):
            return [self.executable, *args[1:]]

        return list(args)

    def full_env(self, overrides: dict[str, str | None] | None) -> dict[str, str]:
        """Generate the full environment for the venv.

        :param overrides: Environment variables to override or unset. Can be ``None`` if
            there are no explicit environment changes. Use ``None`` as a values to unset
            environment variables
        :returns: environment mapping for the venv with overrides applied.
        """

        if overrides:
            env = overrides.copy()
        else:
            env = {}

        old_path = env.get("PATH") or os.environ.get("PATH", "")
        env["PATH"] = os.fspath(self.bin_dir) + (
            os.pathsep + old_path if old_path else ""
        )
        env["VIRTUAL_ENV"] = os.fspath(self.venv_path)

        # Remove PYTHONHOME to avoid conflicts with the venv.
        env.pop("PYTHONHOME", None)

        return env

    def run(self, args: SubprocessArgsT, **kwargs) -> subprocess.CompletedProcess:
        """Run a command in the virtual environment.

        :param args: Command and arguments to run.
        :param kwargs: Additional keyword arguments to pass to subprocess.run.
        :returns: CompletedProcess from the subprocess execution.
        """
        args = self._rewrite_head(list(args))
        user_env = kwargs.pop("env", None)
        kwargs["env"] = self.full_env(user_env)
        return self.tools.subprocess.run(args, **kwargs)

    def Popen(self, args: SubprocessArgsT, **kwargs) -> stdlib_subprocess.Popen:
        """Run a command in the virtual environment using Popen.

        :param args: Command and arguments to run.
        :param kwargs: Additional keyword arguments to pass to subprocess.Popen.
        :returns: A Popen instance for the subprocess execution.
        """
        args = self._rewrite_head(list(args))
        user_env = kwargs.pop("env", None)
        kwargs["env"] = self.full_env(user_env)
        return self.tools.subprocess.Popen(args, **kwargs)

    def check_output(self, args: SubprocessArgsT, **kwargs) -> str:
        """Run a command in the virtual environment and return its output.

        :param args: Command and arguments to run.
        :param kwargs: Additional keyword arguments to pass to subprocess.check_output.
        :returns: String output from the subprocess execution.
        """
        args = self._rewrite_head(list(args))
        user_env = kwargs.pop("env", None)
        kwargs["env"] = self.full_env(user_env)
        return self.tools.subprocess.check_output(args, **kwargs)


class VenvEnvironment:
    """Context manager for creating and managing a venv."""

    def __init__(
        self,
        tools: ToolCache,
        *,
        path: Path,
        recreate: bool = False,
    ):
        self.tools = tools
        self.path = path
        self.recreate = recreate

        self.venv_context = VenvContext(tools=self.tools, venv_path=self.path)

    def __enter__(self):
        if self.recreate:
            with self.tools.console.wait_bar(
                f"Recreating virtual environment ({self.path.name})..."
            ):
                self.venv_context.recreate()
        elif not self.venv_context.exists():
            with self.tools.console.wait_bar(
                f"Creating virtual environment ({self.path.name})..."
            ):
                self.venv_context.create()

        return self.venv_context

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class NoOpVenvContext:
    """Context for wrapping subprocess for no-op venv.

    Provides the same interface as VenvContext but pass calls through to the native
    subprocess.
    """

    def __init__(self, tools, venv_path: Path):
        self.tools = tools
        self.venv_path = venv_path

        self.created = False
        self.marker_path = venv_path / "venv_path"

    def exists(self) -> bool:
        """A no-op env always exists."""
        return True

    def create(self) -> bool:
        """Creates the no-op environment.

        This isn't a strict "create", as the no-op environment will always exist.
        However, the "creation" process will identify if the currently active
        environment has been used previously, and set the `created` flag if the
        environment hasn't been used before. This can then be used to identify "first
        use" behavior, such as installing requirements.
        """
        if not self.marker_path.exists():
            self.marker_path.parent.mkdir(parents=True, exist_ok=True)
            self.marker_path.write_text(sys.executable, encoding="utf-8")
            self.created = True
        try:
            existing_executable = self.marker_path.read_text(encoding="utf-8").strip()
            if existing_executable != sys.executable:
                self.marker_path.write_text(sys.executable, encoding="utf-8")
                self.created = True
        except (OSError, UnicodeDecodeError):
            self.marker_path.write_text(sys.executable, encoding="utf-8")
            self.created = True

    def clean(self) -> None:
        """Remove the virtual environment marker."""
        if self.marker_path.exists():
            self.marker_path.unlink()

    def recreate(self) -> None:
        """Recreate the no-op environment.

        Since the no-op environment always exists, this essentially amounts to ensuring
        that the created flag is set.
        """
        self.clean()
        self.create()

    def run(self, args: SubprocessArgsT, **kwargs) -> subprocess.CompletedProcess:
        """Run command through native subprocess."""
        return self.tools.subprocess.run(args, **kwargs)

    def Popen(self, args: SubprocessArgsT, **kwargs) -> stdlib_subprocess.Popen:
        """Run command through native subprocess.Popen."""
        return self.tools.subprocess.Popen(args, **kwargs)

    def check_output(self, args: SubprocessArgsT, **kwargs) -> str:
        """Run command through native subprocess.check_output."""
        return self.tools.subprocess.check_output(args, **kwargs)


class NoOpEnvironment:
    """A no-op environment that returns a native runner."""

    def __init__(self, tools: ToolCache, *, path: Path, recreate: bool):
        self.tools = tools
        self.path = path
        self.recreate = recreate

        self.noop_context = NoOpVenvContext(tools=tools, venv_path=path)

    def __enter__(self):
        if self.recreate:
            self.noop_context.recreate()
        else:
            self.noop_context.create()

        return self.noop_context

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class VirtualEnvironment(Tool):
    name = "virtual_environment"
    full_name = "Virtual Environment management"

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> "VirtualEnvironment":
        """Make virtual environment available in tool cache."""
        # short circuit since already verified and available
        if hasattr(tools, "virtual_environment"):
            return tools.virtual_environment

        tools.virtual_environment = VirtualEnvironment(tools=tools)
        return tools.virtual_environment

    def create(
        self,
        venv_path: Path,
        *,
        isolated: bool = True,
        recreate: bool = False,
    ) -> VenvEnvironment | NoOpEnvironment:
        """Return a environment context for the requested isolation settings. Creates
        either a virtual environment context or a no-op context.

        :param tools: The tools instance
        :param venv_path: Path for the virtual environment
        :param isolated: If False, return NoOpEnvironment. Default True.
        :param recreate: Whether to recreate existing venv. Default False.
        :returns: A context manager for an environment where code can be executed.
        """
        if not isolated:
            return NoOpEnvironment(
                tools=self.tools,
                path=venv_path,
                recreate=recreate,
            )
        else:
            return VenvEnvironment(
                tools=self.tools,
                path=venv_path,
                recreate=recreate,
            )



ENVIRONMENT_TEST_PARAMS = [
    (None, {}),
    ({"CUSTOM": "value"}, {}),
    (None, {"cwd": "/tmp", "timeout": 30}),
    ({"PATH": "/custom"}, {"cwd": "/tmp", "check": True}),
]


def assert_environment_handling(
    mock_full_env,
    env_override,
    mock_method,
    method_args,
    other_kwargs,
):
    """Assert proper environment handling in subprocess calls."""
    mock_full_env.assert_called_once_with(env_override)

    expected_kwargs = other_kwargs.copy()
    expected_kwargs["env"] = {"FULL": "env"}

    mock_method.assert_called_once_with(method_args, **expected_kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, runtime_checkable

if TYPE_CHECKING:
    from briefcase.config import AppConfig
    from briefcase.console import Console
    from briefcase.integrations.base import ToolCache


@runtime_checkable
class PublishCommandAPI(Protocol):
    """Stable API surface exposed to publication channel plugins.

    This defines the minimal set of attributes and methods that a plugin
    can rely on from the ``command`` parameter passed to ``publish_app()``.
    """

    console: Console
    tools: ToolCache
    dist_path: Path

    def distribution_path(self, app: AppConfig) -> Path: ...


class BasePublicationChannel(ABC):
    """Definition for a plugin that defines a new Briefcase publication channel."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the publication channel."""

    @abstractmethod
    def publish_app(self, app: AppConfig, command: PublishCommandAPI, **options):
        """Publish an application to this channel.

        :param app: The application to publish
        :param command: The publish command instance
        """



from __future__ import annotations

import re
from collections.abc import Collection
from pathlib import Path
from subprocess import CalledProcessError

from packaging.version import Version

from briefcase.exceptions import BriefcaseCommandError, MissingToolError
from briefcase.integrations.base import ManagedTool, ToolCache


class WiX(ManagedTool):
    name = "wix"
    full_name = "WiX"
    supported_host_os: Collection[str] = {"Windows"}

    # WARNING: version 6 and later have licensing issues: see
    # https://github.com/beeware/briefcase/issues/1185.
    version = Version("5.0.2")

    def __init__(self, tools: ToolCache, wix_home: Path | None = None):
        """Create a wrapper around a WiX install.

        :param tools: ToolCache of available tools.
        :param wix_home: The path of the WiX installation.
        """
        super().__init__(tools=tools)
        if wix_home:
            self.wix_home = wix_home
        else:
            self.wix_home = tools.base_path / "wix"

    @property
    def download_url(self) -> str:
        return (
            f"https://github.com/wixtoolset/wix/releases/download/v{self.version}/"
            f"wix-cli-x64.msi"
        )

    @property
    def wix_exe(self) -> Path:
        return (
            self.wix_home
            / f"PFiles64/WiX Toolset v{self.version.major}.{self.version.minor}/bin"
            / "wix.exe"
        )

    def ext_path(self, name: str) -> Path:
        return (
            self.wix_home
            / f"CFiles64/WixToolset/extensions/WixToolset.{name}.wixext/{self.version}"
            / f"wixext{self.version.major}/WixToolset.{name}.wixext.dll"
        )

    @classmethod
    def verify_install(cls, tools: ToolCache, install: bool = True, **kwargs) -> WiX:
        """Verify that there is a WiX install available.

        WiX is a small tool, and there's a close relationship between the WiX version
        and the template syntax, so we always use a Briefcase-managed copy, and upgrade
        it automatically.

        :param tools: ToolCache of available tools
        :param install: Should WiX be installed if it is not found?
        """
        # short circuit since already verified and available
        if hasattr(tools, "wix"):
            return tools.wix

        wix = WiX(tools)
        if not wix.exists():
            if install:
                tools.console.info(
                    "The WiX toolset was not found; downloading and installing...",
                    prefix=cls.name,
                )
                wix.install()
            else:
                raise MissingToolError("WiX")
        else:
            try:
                # The string returned by --version may include "+" followed by a
                # commit ID; ignore this.
                installed_version = re.sub(
                    r"\+.*",
                    "",
                    tools.subprocess.check_output([wix.wix_exe, "--version"]),
                ).strip()
            except (OSError, CalledProcessError) as e:
                installed_version = None
                tools.console.error(
                    f"The WiX toolset is unusable ({type(e).__name__}: {e})"
                )

            if installed_version != str(wix.version):
                if installed_version is not None:
                    tools.console.info(
                        f"The WiX toolset is an unsupported version "
                        f"({installed_version})"
                    )
                if install:
                    wix.upgrade()
                else:
                    raise MissingToolError("WiX")

        tools.console.debug(f"Using WiX at {wix.wix_home}")
        tools.wix = wix
        return wix

    def exists(self) -> bool:
        return self.wix_exe.is_file()

    @property
    def managed_install(self) -> bool:
        return True

    def install(self):
        """Download and install WiX."""
        wix_msi_path = self.tools.file.download(
            url=self.download_url,
            download_path=self.tools.base_path,
            role="WiX",
        )

        try:
            with self.tools.console.wait_bar("Installing WiX..."):
                self.tools.subprocess.run(
                    [
                        "msiexec",
                        "/a",  # Unpack the MSI into individual files
                        wix_msi_path,
                        "/qn",  # Disable GUI interaction
                        f"TARGETDIR={self.wix_home}",
                    ],
                    check=True,
                )
        except CalledProcessError as e:
            raise BriefcaseCommandError(
                f"""\
Unable to unpack WiX MSI file. The download may have been
interrupted or corrupted.

Delete {wix_msi_path} and run briefcase again.
"""
            ) from e

        wix_msi_path.unlink()

    def uninstall(self):
        """Uninstall WiX."""
        with self.tools.console.wait_bar("Removing old WiX install..."):
            self.tools.shutil.rmtree(self.wix_home)



from briefcase.config import BaseConfig


def test_update():
    # Expect update() to overwrite an existing field
    config = BaseConfig()
    config.requires = ["newpackage=1.1"]
    new_requirements = ["newpackage=1.2"]
    config.update({"requires": new_requirements})
    assert config.requires == new_requirements


def test_setdefault():
    config = BaseConfig()
    # Expect setdefault() to add new field since it does not exist
    result = config.setdefault("requires", [])
    assert result == config.requires
    assert config.requires == []

    # Expect setdefault() to not overwrite existing field
    new_requirements = config.requires = ["newpackage=1.2"]
    result = config.setdefault("requires", [])
    assert result == config.requires
    assert config.requires == new_requirements



from __future__ import annotations

import argparse

from briefcase.channels import get_publication_channels
from briefcase.channels.base import BasePublicationChannel
from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError

from .base import BaseCommand, full_options


class PublishCommand(BaseCommand):
    command = "publish"
    description = "Publish an app to a distribution channel."

    @property
    def packaging_formats(self):
        return [self.output_format]

    @property
    def default_packaging_format(self):
        return self.output_format

    def _get_channels(self) -> dict[str, type[BasePublicationChannel]]:
        """Discover available publication channels for this platform/format."""
        return get_publication_channels(self.platform, self.output_format)

    def add_options(self, parser: argparse.ArgumentParser) -> None:
        channels = self._get_channels()
        channel_names = sorted(channels.keys())

        parser.add_argument(
            "-a",
            "--app",
            dest="app_name",
            help="Name of the app to publish (if multiple apps exist in the project)",
            default=argparse.SUPPRESS,
        )

        parser.add_argument(
            "-u",
            "--update",
            action="store_true",
            help="Update the app before publishing",
        )
        parser.add_argument(
            "-p",
            "--packaging-format",
            dest="packaging_format",
            help="Packaging format to publish",
            default=self.default_packaging_format,
            choices=self.packaging_formats,
        )
        parser.add_argument(
            "-c",
            "--channel",
            choices=channel_names or None,
            default=None,
            help="The channel to publish to",
        )

    def _publish_app(
        self,
        app: AppConfig,
        update: bool,
        packaging_format: str,
        channel: BasePublicationChannel,
        **options,
    ) -> dict | None:
        """Internal method to publish a single app.

        :param app: The application to publish
        :param update: Should the application be updated (and rebuilt) first?
        :param packaging_format: The format of the packaging artefact to create.
        :param channel: The resolved BasePublicationChannel instance
        """
        state = None

        # Annotate the packaging format onto the app
        app.packaging_format = packaging_format

        if update or not self.distribution_path(app).exists():
            state = self.package_command(
                app,
                update=update,
                packaging_format=packaging_format,
                **options,
            )

        self.verify_app(app)

        state = channel.publish_app(app, command=self, **full_options(state, options))

        return state

    def __call__(
        self,
        app: AppConfig | None = None,
        app_name: str | None = None,
        update: bool | None = False,
        packaging_format: str | None = None,
        channel: str | None = None,
        **options,
    ) -> dict | None:
        apps_to_publish = self.resolve_apps(app=app, app_name=app_name)

        # Verify that at least one publication channel is available.
        channels = self._get_channels()
        if not channels:
            raise BriefcaseCommandError(
                f"No publication channels are available for "
                f"{self.platform} {self.output_format}.\n\n"
                f"Install a publication channel plugin and try again."
            )

        if channel is None:
            if len(channels) == 1:
                resolved_channel = next(iter(channels.values()))()
            else:
                raise BriefcaseCommandError(
                    f"Multiple publication channels are available for "
                    f"{self.platform} {self.output_format}: "
                    f"{', '.join(sorted(channels))}.\n\n"
                    f"Specify a channel with --channel."
                )
        else:
            resolved_channel = channels[channel]()

        # Confirm host compatibility, that all required tools are available,
        # and that all app configurations are finalized.
        self.finalize(apps=apps_to_publish.values())

        # Publish all apps to the selected channel.
        state = None
        for _, app_obj in sorted(apps_to_publish.items()):
            state = self._publish_app(
                app_obj,
                update=update,
                packaging_format=packaging_format or self.default_packaging_format,
                channel=resolved_channel,
                **full_options(state, options),
            )

        return state



import re
import sys
from pathlib import Path

import debugpy

from briefcase_debugger.config import DebuggerConfig


def find_first_matching_path(regex: str) -> str:
    """Returns the first element of sys.paths that matches regex, otherwise None."""
    for path in sys.path:
        if re.search(regex, path):
            return path
    raise ValueError(f"No sys.path entry matches regex '{regex}'")


def load_path_mappings(config: DebuggerConfig, verbose: bool) -> list[tuple[str, str]]:
    app_path_mappings = config.get("app_path_mappings", None)
    app_packages_path_mappings = config.get("app_packages_path_mappings", None)

    mappings_list = []
    if app_path_mappings:
        device_app_folder = find_first_matching_path(
            app_path_mappings["device_sys_path_regex"]
        )
        for app_subfolder_device, app_subfolder_host in zip(
            app_path_mappings["device_subfolders"],
            app_path_mappings["host_folders"],
            strict=False,
        ):
            mappings_list.append(
                (
                    app_subfolder_host,
                    str(Path(device_app_folder) / app_subfolder_device),
                )
            )
    if app_packages_path_mappings:
        device_app_packages_folder = find_first_matching_path(
            app_packages_path_mappings["sys_path_regex"]
        )
        mappings_list.append(
            (
                app_packages_path_mappings["host_folder"],
                str(Path(device_app_packages_folder)),
            )
        )

    if verbose:
        print("Extracted path mappings:")
        for idx, mapping in enumerate(mappings_list):
            print(f"[{idx}] host =   {mapping[0]}")
            print(f"[{idx}] device = {mapping[1]}")

    return mappings_list


def start_debugpy(config: DebuggerConfig, verbose: bool):
    host = config["host"]
    port = config["port"]
    path_mappings = load_path_mappings(config, verbose)

    # Starting remote debugger...
    print(f"Starting debugpy in server mode at {host}:{port}...")
    debugpy.listen((host, port), in_process_debug_adapter=True)

    if verbose:
        # pydevd is dynamically loaded and only available after debugpy is started
        import pydevd

        pydevd.DebugInfoHolder.DEBUG_TRACE_LEVEL = 3

    if len(path_mappings) > 0:
        if verbose:
            print("Adding path mappings...")

        # pydevd is dynamically loaded and only available after a debugger has connected
        import pydevd_file_utils

        pydevd_file_utils.setup_client_server_paths(path_mappings)

    print("The debugpy server started. Waiting for debugger to attach...")
    print(
        f"""
To connect to debugpy using VS Code add the following configuration to '.vscode/launch.json':
{{
    "version": "0.2.0",
    "configurations": [
        {{
            "name": "Briefcase: Attach (Connect)",
            "type": "debugpy",
            "request": "attach",
            "connect": {{
                "host": "{host}",
                "port": {port}
            }},
            "justMyCode": false
        }}
    ]
}}

For more information see: https://briefcase.beeware.org/en/stable/how-to/debugging/vscode/#bundled-app
"""
    )
    debugpy.wait_for_client()

    print("Debugger attached.")
    print("-" * 75)



from textwrap import dedent

import pytest

from ....utils import create_file


# Windows' AppConfig requires attribute 'packaging_format'
@pytest.fixture
def first_app_templated(first_app_config, tmp_path):
    bundle_path = tmp_path / "base_path/build/first-app/windows/app"

    # Create the stub binary
    create_file(bundle_path / "src/Stub.exe", "Stub binary")

    # Create briefcase.tomml
    create_file(
        bundle_path / "briefcase.toml",
        dedent(
            """\
            [paths]
            extras_path = "custom_extras"
            """
        ),
    )

    return first_app_config



from pathlib import Path

from .conftest import DummyCommand


def test_briefcase_required_python_version(base_command):
    assert base_command.briefcase_required_python_version == (3, 10)


def test_bundle_path(base_command, my_app, tmp_path):
    bundle_path = base_command.bundle_path(my_app)

    assert bundle_path == tmp_path / "base_path/build/my-app/tester/dummy"


def test_package_path(base_command, my_app, tmp_path):
    package_path = base_command.package_path(my_app)
    bundle_package_path = base_command.bundle_package_path(my_app)

    assert package_path == tmp_path / "base_path/build/my-app/tester/dummy/src/package"
    assert (
        bundle_package_path
        == tmp_path / "base_path/build/my-app/tester/dummy/src/package"
    )


def test_external_package_path(base_command, my_app, tmp_path):
    my_app.external_package_path = "path/to/package"

    package_path = base_command.package_path(my_app)
    bundle_package_path = base_command.bundle_package_path(my_app)

    assert package_path == (Path.cwd() / "path/to/package")
    assert (
        bundle_package_path
        == tmp_path / "base_path/build/my-app/tester/dummy/src/package"
    )


def test_package_executable_path(base_command, my_app, tmp_path):
    package_executable_path = base_command.package_executable_path(my_app)
    bundle_package_executable_path = base_command.bundle_package_executable_path(my_app)

    assert package_executable_path == "internal/my-app.exe"
    assert bundle_package_executable_path == "internal/my-app.exe"


def test_external_package_executable_path(base_command, my_app, tmp_path):
    my_app.external_package_executable_path = "alternate/the_app.exe"

    package_executable_path = base_command.package_executable_path(my_app)
    bundle_package_executable_path = base_command.bundle_package_executable_path(my_app)

    assert package_executable_path == "alternate/the_app.exe"
    assert bundle_package_executable_path == "internal/my-app.exe"


def test_create_command(base_command):
    # Check for a property of the created command class.
    assert base_command.create_command.description == "Test Create"


def test_update_command(base_command):
    # Check for a property of the created command class.
    assert base_command.update_command.description == "Test Update"


def test_build_command(base_command):
    # Check for a property of the created command class.
    assert base_command.build_command.description == "Test Build"


def test_run_command(base_command):
    # Check for a property of the created command class.
    assert base_command.run_command.description == "Test Run"


def test_package_command(base_command):
    # Check for a property of the created command class.
    assert base_command.package_command.description == "Test Package"


def test_publish_command(base_command):
    # Check for a property of the created command class.
    assert base_command.publish_command.description == "Test Publish"


def test_command_state_transferred(dummy_console, tmp_path):
    """Command state is transferred to created subcommands."""
    command = DummyCommand(console=dummy_console, base_path=tmp_path)
    command.tools.console.input_enabled = False

    # Check the enabled state of subcommands
    assert not command.create_command.console.input_enabled
    assert command.create_command.console is command.tools.console
    assert command.create_command.tools is command.tools
    assert command.create_command.is_clone is True

    assert not command.update_command.console.input_enabled
    assert command.update_command.console is command.tools.console
    assert command.update_command.tools is command.tools
    assert command.update_command.is_clone is True

    assert not command.build_command.console.input_enabled
    assert command.build_command.console is command.tools.console
    assert command.build_command.tools is command.tools
    assert command.build_command.is_clone is True

    assert not command.run_command.console.input_enabled
    assert command.run_command.console is command.tools.console
    assert command.run_command.tools is command.tools
    assert command.run_command.is_clone is True

    assert not command.package_command.console.input_enabled
    assert command.package_command.console is command.tools.console
    assert command.package_command.tools is command.tools
    assert command.package_command.is_clone is True

    assert not command.publish_command.console.input_enabled
    assert command.publish_command.console is command.tools.console
    assert command.publish_command.tools is command.tools
    assert command.publish_command.is_clone is True



def test_exists_always_returns_true(noop_context):
    """Exists always returns True for no-op context, regardless of marker state."""
    # Test when marker doesn't exist
    assert not noop_context.marker_path.exists()
    assert noop_context.exists() is True

    # Test when marker does exist
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    noop_context.marker_path.write_text("something", encoding="utf-8")
    assert noop_context.exists() is True



# The package command inherits most of its behavior from the common base
# implementation. Do a surface-level verification here, but the app
# tests provide the actual test coverage.
from unittest import mock

import pytest

from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.wix import WiX
from briefcase.platforms.windows.visualstudio import WindowsVisualStudioPackageCommand

from ....integrations.wix.conftest import WIX_EXE_PATH, WIX_UI_PATH


@pytest.fixture
def package_command(dummy_console, tmp_path):
    command = WindowsVisualStudioPackageCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.wix = WiX(command.tools, wix_home=tmp_path / "wix")
    return command


def test_package_msi(package_command, first_app_config, tmp_path):
    """A Windows app can be packaged as an MSI."""

    package_command.package_app(first_app_config)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix" / WIX_EXE_PATH,
                "build",
                "-ext",
                tmp_path / "wix" / WIX_UI_PATH,
                "-arch",
                "x64",
                "first-app.wxs",
                "-loc",
                "unicode.wxl",
                "-pdbtype",
                "none",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/visualstudio",
        ),
    ]



import pytest

from briefcase.config import parse_boolean


@pytest.mark.parametrize(
    "user_input_val",
    ["true", "TRUE", "t", "T", "yes", "YES", "y", "Y", "1", "on", "ON"],
)
def test_true_values(user_input_val):
    assert parse_boolean(user_input_val) is True


@pytest.mark.parametrize(
    "user_input_val",
    ["false", "FALSE", "f", "F", "no", "NO", "n", "N", "0", "off", "OFF"],
)
def test_false_values(user_input_val):
    assert parse_boolean(user_input_val) is False


@pytest.mark.parametrize("user_input_val", ["maybe", "2", "", "help"])
def test_invalid_values(user_input_val):
    with pytest.raises(ValueError, match="Invalid boolean value"):
        parse_boolean(user_input_val)


@pytest.mark.parametrize(
    ("user_input_val", "expected"),
    [("         yEs ", True), ("nO          ", False), ("    YEs", True)],
)
def test_whitespace(user_input_val, expected):
    assert parse_boolean(user_input_val) is expected



import re
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.virtual_environment import VenvContext


def test_update_core_tools_success(mock_tools, venv_path):
    """update_core_tools succeeds."""
    context = VenvContext(mock_tools, venv_path)
    context.run = MagicMock()
    context.update_core_tools()
    context.run.assert_called_once_with(
        [context.executable, "-m", "pip", "install", "-U", "pip"],
        check=True,
    )


def test_update_core_tools_failure(mock_tools, venv_path):
    """Update core tools raise BriefcaseCommandError if subprocess fails."""
    context = VenvContext(mock_tools, venv_path)
    test_exception = RuntimeError("pip install failed")
    context.run = MagicMock(side_effect=test_exception)
    escaped_venv_path = re.escape(
        str(venv_path)
    )  # Escape backslashes for regex match, windows backslashes
    with pytest.raises(
        BriefcaseCommandError,
        match=f"Failed to update core tooling for {escaped_venv_path}",
    ):
        context.update_core_tools()
    context.run.assert_called_once_with(
        [context.executable, "-m", "pip", "install", "-U", "pip"],
        check=True,
    )



import json
from importlib import metadata
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from briefcase.debuggers import (
    DebugpyDebugger,
    PdbDebugger,
    get_debugger,
    get_debuggers,
)
from briefcase.debuggers.base import DebuggerConnectionMode, _is_editable_pep610
from briefcase.exceptions import BriefcaseCommandError


class DummyDist:
    def __init__(self, direct_url):
        self._direct_url = direct_url

    def read_text(self, name):
        return self._direct_url if name == "direct_url.json" else None


@pytest.mark.parametrize(
    ("direct_url", "is_editable"),
    [
        (json.dumps({"dir_info": {"editable": True}}), True),  # editable
        (json.dumps({"dir_info": {"editable": False}}), False),  # not editable
        (json.dumps({}), False),  # missing dir_info
        (None, False),  # missing direct_url.json
        ("not-json", False),  # invalid JSON
    ],
)
def test_is_editable_pep610(monkeypatch, direct_url, is_editable):
    """Detection of editable installs via PEP 610 direct_url.json works."""
    monkeypatch.setattr(metadata, "distribution", lambda name: DummyDist(direct_url))
    assert _is_editable_pep610("briefcase") is is_editable


def test_is_editable_pep610_package_not_found(monkeypatch):
    """Detection of editable install throws an Error if package is not found."""

    def raise_not_found(name):
        raise metadata.PackageNotFoundError

    monkeypatch.setattr(metadata, "distribution", raise_not_found)
    with pytest.raises(metadata.PackageNotFoundError):
        _is_editable_pep610("briefcase")


def test_get_debuggers():
    """Builtin debuggers are available."""
    debuggers = get_debuggers()
    assert isinstance(debuggers, dict)
    assert debuggers["pdb"] is PdbDebugger
    assert debuggers["pdb"]().name == "pdb"
    assert debuggers["debugpy"] is DebugpyDebugger
    assert debuggers["debugpy"]().name == "debugpy"


def test_get_debugger():
    """Debugger can be retrieved by name."""
    assert isinstance(get_debugger("pdb"), PdbDebugger)
    assert isinstance(get_debugger("debugpy"), DebugpyDebugger)

    # Test with an unknown debugger name
    with pytest.raises(BriefcaseCommandError, match="Unknown debugger: unknown"):
        get_debugger("unknown")


@pytest.mark.parametrize(
    ("debugger_name", "expected_class", "connection_mode"),
    [
        (
            "pdb",
            PdbDebugger,
            DebuggerConnectionMode.SERVER,
        ),
        (
            "debugpy",
            DebugpyDebugger,
            DebuggerConnectionMode.SERVER,
        ),
    ],
)
def test_debugger(debugger_name, expected_class, connection_mode, monkeypatch):
    """Debugger uses correct connection mode and support package."""
    monkeypatch.setattr("briefcase.debuggers.base.IS_EDITABLE", False)

    debugger = get_debugger(debugger_name)
    assert isinstance(debugger, expected_class)
    assert debugger.connection_mode == connection_mode
    assert f"briefcase-debugger[{debugger_name}]" in debugger.debugger_support_pkg


@pytest.mark.parametrize(
    "debugger_name",
    ["pdb", "debugpy"],
)
def test_debugger_editable(debugger_name, monkeypatch):
    """Debugger support package is local path in editable briefcase install."""
    with TemporaryDirectory() as tmp_path:
        tmp_path = Path(tmp_path)
        (tmp_path / "debugger").mkdir(parents=True, exist_ok=True)
        monkeypatch.setattr("briefcase.debuggers.base.IS_EDITABLE", True)
        monkeypatch.setattr("briefcase.debuggers.base.REPO_ROOT", tmp_path)

        debugger = get_debugger(debugger_name)
        assert (
            str(tmp_path / f"debugger[{debugger_name}]")
            == debugger.debugger_support_pkg
        )


@pytest.mark.parametrize(
    "debugger_name",
    ["pdb", "debugpy"],
)
def test_debugger_editable_path_not_found(debugger_name, monkeypatch):
    """Debugger support package is not the local path when path is not available."""
    with TemporaryDirectory() as tmp_path:
        tmp_path = Path(tmp_path)
        monkeypatch.setattr("briefcase.debuggers.base.IS_EDITABLE", True)
        monkeypatch.setattr("briefcase.debuggers.base.REPO_ROOT", tmp_path)

        debugger = get_debugger(debugger_name)
        assert f"briefcase-debugger[{debugger_name}]==" in debugger.debugger_support_pkg



import pytest

from ....utils import create_file, create_wheel


@pytest.fixture
def first_app_generated(first_app_config, tmp_path):
    # Create the briefcase.toml file
    bundle_path = tmp_path / "base_path/build/first-app/web/static"
    create_file(
        bundle_path / "briefcase.toml",
        """
[paths]
app_path="app"
app_requirements_path="requirements.txt"
""",
    )

    # Create index.html with insert markers
    create_file(
        bundle_path / "www/index.html",
        """<!doctype html>
<html>
  <head>
    <!--@@ head:start @@-->
    <!--@@ head:end @@-->
    <!--@@ head-python:start @@-->
    <!--@@ head-python:end @@-->
  </head>
  <body>
    <div id="briefcase-splash"></div>

    <!--@@ body-start:start @@-->
    <!--@@ body-start:end @@-->

    <!--@@ body-python:start @@-->
    <!--@@ body-python:end @@-->

    <!--@@ body-end:start @@-->
    <!--@@ body-end:end @@-->
  </body>
</html>
""",
    )

    # Create the initial style.css with CSS insert markers
    create_file(
        bundle_path / "www/static/css/style.css",
        """
#pyconsole {
  display: None;
}
/*******************************************************************
******************** Wheel contributed styles ********************/
/*@@ css:start @@*/
/*@@ css:end @@*/

""",
    )

    # Create an empty wheels folder
    (bundle_path / "www/static/wheels").mkdir(parents=True)

    return first_app_config


@pytest.fixture
def first_app_built(first_app_generated, tmp_path):
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Create an app wheel
    create_wheel(bundle_path / "www/static/wheels")

    return first_app_generated



import importlib
import json
import os
import sys
from unittest.mock import MagicMock

import briefcase_debugger


def test_import_for_code_coverage(monkeypatch, capsys):
    """Get 100% code coverage."""
    # The module `briefcase_debugger` is already imported through the .pth
    # file. Code executed during .pth files are not covered by coverage.py.
    # So we need to reload the module to get a 100% code coverage.
    importlib.reload(importlib.import_module("briefcase_debugger"))


def test_unknown_debugger(monkeypatch, capsys):
    """An unknown debugger raises an error and stops the application."""
    os_environ = {}
    os_environ["BRIEFCASE_DEBUGGER"] = json.dumps(
        {
            "debugger": "unknown",
            "host": "somehost",
            "port": 9999,
            "host_os": "Windows",
        }
    )
    monkeypatch.setattr(os, "environ", os_environ)

    fake_sys_exit = MagicMock()
    monkeypatch.setattr(sys, "exit", fake_sys_exit)

    briefcase_debugger.start_remote_debugger()

    fake_sys_exit.assert_called_once_with(-1)

    captured = capsys.readouterr()
    assert "Unknown debugger" in captured.out



from __future__ import annotations

import itertools
import os
import shutil
import ssl
import sys
import tempfile
from collections.abc import Iterable, Sequence
from contextlib import suppress
from email.message import Message
from pathlib import Path

import httpx
import truststore

from briefcase.exceptions import (
    BadNetworkResourceError,
    MissingNetworkResourceError,
    NetworkFailure,
)
from briefcase.integrations.base import Tool, ToolCache


class File(Tool):
    name = "file"
    full_name = "File"

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> File:
        """Make File available in tool cache."""
        # short circuit since already verified and available
        if hasattr(tools, "file"):
            return tools.file

        tools.file = File(tools=tools)
        return tools.file

    @classmethod
    def sorted_depth_first(cls, paths: Sequence[Path]) -> Iterable[Path]:
        """Sort a list of paths, so that they appear in lexical order, with
        subdirectories of a folder sorting before files in the same directory.

        :param paths: The list of paths to sort.
        :returns: The sorted list of paths
        """
        # The sort key for a path is a triple - the parent of the path; whether the path
        # is a directory or not; and finally, the actual path. We then sort this in
        # reverse order.
        #
        # Sorting by the parent of the path first (in reverse order) guarantees that
        # long paths are sorted first; so a file in a folder will come *after* the
        # folder it is in.
        #
        # The second term (parent.is_dir()) guarantees that subfolders in a folder are
        # found before files in the same folder (because in reverse order, booleans sort
        # True before False).
        #
        # Lastly, we sort on the actual path itself. This provide a guaranteed lexical
        # ordering inside any given folder. This isn't strictly required, but it's
        # helpful for testing and reproducibility that the sort order is reliable and
        # repeatable.
        return sorted(paths, key=lambda p: (p.parent, p.is_dir(), p), reverse=True)

    @classmethod
    def sorted_depth_first_groups(
        cls,
        paths: Sequence[Path],
    ) -> Iterable[Iterable[Path]]:
        """Convert a list of paths into a collection of groups that are all at the same
        "level", grouped depth-first.

        Subfolders in a folder will be in a separate group, returned *before* files in
        the same folder.

        :param paths: The list of paths return grouped.
        :returns: A generator returning a iterable groups of paths that are all at the
            same sorting level.
        """
        # The sort function guarantees depth first ordering, with folders before files
        # at the same level.
        #
        # The grouping key is based on the parent of the path (so that all objects in
        # the same folder group together), with an additional discriminator to ensure
        # that subfolders are in a different group to files.
        #
        # itertools.groupby guarantees a new group whenever the key changes; and the
        # input sort order guarantees that directories come before files, so we get
        # our desired sort grouping and ordering.
        return (
            group_paths
            for _, group_paths in itertools.groupby(
                cls.sorted_depth_first(paths),
                lambda path: (path.parent, path.is_dir()),
            )
        )

    @property
    def ssl_context(self):
        """The SSL context to use for downloads."""
        try:
            return self._ssl_context
        except AttributeError:
            self._ssl_context = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

            return self._ssl_context

    def is_archive(self, filename: str | os.PathLike) -> bool:
        """Can a file be unpacked via `shutil.unpack_archive()`?

        The evaluation is based purely on the name of the file. A more robust
        implementation may actually interrogate the file itself.

        Notably, the types of archives that can be unpacked will vary based on the types
        of compression the current Python supports. So, for example, if LZMA is not
        available, then tar.xz archives cannot be unpacked.

        :param filename: path to file to evaluate as an archive
        """
        filename = Path(filename)
        file_extensions = {
            # captures extensions like .tar.gz, .tar.bz2, etc.
            "".join(filename.suffixes[-2:]),
            # as well as .tar, .zip, etc.
            filename.suffix,
        }
        return not file_extensions.isdisjoint(self.supported_archive_extensions)

    @property
    def supported_archive_extensions(self) -> set[str]:
        return {
            extension
            for archive_format in shutil.get_unpack_formats()
            for extension in archive_format[1]
        }

    def unpack_archive(
        self,
        filename: str | os.PathLike,
        extract_dir: str | os.PathLike,
        **kwargs,
    ):
        """Unpack an archive file in to a destination directory.

        Additional protections for unpacking tar files were introduced in Python 3.12.
        Since tarballs can contain anything valid in a UNIX file system, these
        protections prevent unpacking potentially dangerous files. This behavior will be
        the default in Python 3.14. However, the protections can only be enabled for tar
        files...not zip files.

        :param filename: File path for the archive
        :param extract_dir: Target file path for where to unpack archive
        :param kwargs: additional arguments for shutil.unpack_archive
        """
        is_zip = Path(filename).suffix == ".zip"
        if sys.version_info >= (3, 12):  # pragma: no-cover-if-lt-py312
            unpack_kwargs = {"filter": "data"} if not is_zip else {}
        else:  # pragma: no-cover-if-gte-py312
            unpack_kwargs = {}

        self.tools.shutil.unpack_archive(
            filename=filename,
            extract_dir=extract_dir,
            **{
                **unpack_kwargs,
                **kwargs,
            },
        )

    def download(self, url: str, download_path: Path, role: str | None = None) -> Path:
        """Download a given URL, caching it. If it has already been downloaded, return
        the value that has been cached.

        This is a utility method used to obtain assets used by the installation process.
        The cached filename will be the filename portion of the URL, appended to the
        download path.

        :param url: The URL to download
        :param download_path: The path to the download cache folder. This path will be
            created if it doesn't exist.
        :param role: A string describing the role played by the file being downloaded;
            used to construct log and error messages. Should be able to fit into the
            sentence "Error downloading {role}".
        :returns: The filename of the downloaded (or cached) file.
        """
        download_path.mkdir(parents=True, exist_ok=True)
        filename: Path | None = None
        try:
            with self.tools.httpx.stream(
                "GET",
                url,
                follow_redirects=True,
                verify=self.ssl_context,
            ) as response:
                if response.status_code == 404:
                    raise MissingNetworkResourceError(url=url)
                elif response.status_code != 200:
                    raise BadNetworkResourceError(
                        url=url, status_code=response.status_code
                    )

                # The initial URL might go through URL redirects, so we need the *final*
                # response. We look at either the `Content-Disposition` header or the
                # final URL to extract the cache filename.
                cache_full_name = response.url.path
                header_value = response.headers.get("Content-Disposition")
                if header_value:
                    # Httpx does not provide a way to parse RFC6266 headers.
                    # The cgi module *did* have a way to parse these headers, but
                    # it was deprecated as part of PEP594. PEP594 recommends
                    # using the email.message module to parse these headers as they
                    # are near identical format.
                    # See also:
                    # * https://tools.ietf.org/html/rfc6266
                    # * https://peps.python.org/pep-0594/#cgi
                    msg = Message()
                    msg["Content-Disposition"] = header_value
                    filename = msg.get_filename()
                    if filename:
                        cache_full_name = filename
                cache_name = cache_full_name.split("/")[-1]
                filename = download_path / cache_name

                if filename.exists():
                    self.tools.console.info(f"{cache_name} already downloaded")
                else:
                    self.tools.console.info(f"Downloading {cache_name}...")
                    self._fetch_and_write_content(response, filename)
        except httpx.RequestError as e:
            if role:
                description = role
            else:
                description = filename.name if filename else url

            if isinstance(e, httpx.TooManyRedirects):
                # httpx, unlike requests, will not follow redirects indefinitely and
                # defaults to 20 redirects before calling it quits. If the download
                # attempt exceeds 20 redirects, Briefcase probably needs to re-evaluate
                # the URLs it is using for that download and ideally find a starting
                # point that won't have so many redirects.
                hint = (
                    "exceeded redirects when downloading the file.\n\n"
                    "Please report this as a bug to Briefcase."
                )
            elif isinstance(e, httpx.DecodingError):
                hint = "the server sent a malformed response."
            elif isinstance(e, httpx.ConnectError):
                try:
                    # It's a little difficult to verify exactly what might cause httpx
                    # to raise a ConnectError, but `__context__.__context__` should be
                    # an SSLCertVerificationError if there's a certificate problem.
                    # Catch that case, and print the raw exception in other cases.
                    context = e.__context__.__context__
                except AttributeError:
                    context = None

                if isinstance(context, ssl.SSLCertVerificationError):
                    hint = (
                        "a connection to the server could not be established due to "
                        "a certificate verification problem.\n\n"
                        "This probably indicates an issue with a proxy configuration "
                        "on your computer."
                    )
                else:
                    hint = (
                        "a connection to the server could not be established.\n\n"
                        f"The reported cause of the problem was {e}"
                    )
            else:
                # httpx.TransportError
                # Use the default hint for generic network communication errors
                hint = (
                    "is your computer offline?\n\n"
                    f"The reported cause of the problem was {e}"
                )

            raise NetworkFailure(
                f"download {description}",
                hint,
            ) from e

        return filename

    def _fetch_and_write_content(self, response: httpx.Response, filename: Path):
        """Write the content from the httpx Response to file.

        The data is initially written in to a temporary file in the Briefcase
        cache. This avoids partially downloaded files masquerading as complete
        downloads in later Briefcase runs. The temporary file is only moved
        to ``filename`` if the download is successful; otherwise, it is deleted.

        :param response: ``httpx.Response``
        :param filename: full filesystem path to save data
        """
        # `temp_file` is used in the `finally` block, so make sure it's assigned
        # before the `try`.
        temp_file = tempfile.NamedTemporaryFile(  # noqa: SIM115 (use context manager)
            dir=filename.parent,
            prefix=f"{filename.name}.",
            suffix=".download",
            delete=False,
        )
        try:
            with temp_file:
                total = response.headers.get("content-length")
                if total is None:
                    response.read()
                    temp_file.write(response.content)
                else:
                    progress_bar = self.tools.console.progress_bar()
                    task_id = progress_bar.add_task("Downloader", total=int(total))
                    with progress_bar:
                        for data in response.iter_bytes(chunk_size=1024 * 1024):
                            temp_file.write(data)
                            progress_bar.update(task_id, advance=len(data))

            # This file move short circuits to a file rename when the source and
            # destination are on the same filesystem; therefore, it should complete
            # quite quickly even for large files.
            self.tools.shutil.move(temp_file.name, filename)
            # Temporary files are created with only read/write permissions for the
            # file's owner (i.e. 600); to match the behavior of file creation using
            # ``open(..., "w")``, the downloaded file's permissions are updated for
            # the group and world to have read/write permissions as well. Finally,
            # (as with ``open()``) the system's umask is respected. The current
            # umask is only available as the return value to updating the umask...
            # Updating the umask affects the current process, including all threads.
            # A umask value represents permissions that should be denied; so, 022
            # denies write permissions to the group and world. A 022 umask is a
            # common default among supporting systems.
            os.umask(current_umask := os.umask(0o022))
            # The umask is applied by inverting it and bitwise ANDing the default
            # permissions...thus masking out permissions that should be denied.
            self.tools.os.chmod(filename, 0o666 & ~current_umask)
        finally:
            # Ensure the temporary file is deleted; this file may still
            # exist if the download fails or the user sends CTRL+C.
            with suppress(FileNotFoundError):
                self.tools.os.remove(temp_file.name)



from __future__ import annotations

from typing import TYPE_CHECKING

from briefcase.channels.base import BasePublicationChannel
from briefcase.exceptions import BriefcaseCommandError

if TYPE_CHECKING:
    from briefcase.channels.base import PublishCommandAPI
    from briefcase.config import AppConfig


class AppStorePublicationChannel(BasePublicationChannel):
    """Placeholder for iOS App Store publication channel."""

    @property
    def name(self) -> str:
        return "appstore"

    def publish_app(self, app: AppConfig, command: PublishCommandAPI, **options):
        raise BriefcaseCommandError(
            "Publishing to the iOS App Store is not yet implemented.\n"
            "\n"
            "See https://github.com/beeware/briefcase/issues/2697 for progress."
        )



import pytest

from briefcase.exceptions import BriefcaseCommandError

from .conftest import _make_channel_class


def test_channel_classes_are_instantiated(publish_command, first_app):
    """Channel classes returned by _get_channels are instantiated before use."""
    instantiated = []

    SpyChannel = _make_channel_class("spy")
    original_init = SpyChannel.__init__

    def tracking_init(self):
        instantiated.append(True)
        original_init(self)

    SpyChannel.__init__ = tracking_init

    # _get_channels returns classes, matching the real entry_points().load() behavior
    publish_command._get_channels = lambda: {"spy": SpyChannel}
    publish_command.apps = {"first": first_app}

    publish_command(channel="spy")

    assert len(instantiated) == 1
    assert ("publish", "first", "spy", {}) in publish_command.actions


def test_no_channels(publish_command, first_app):
    """If no publication channels are available, an error is raised."""
    publish_command._get_channels = dict
    publish_command.apps = {"first": first_app}

    options, _ = publish_command.parse_options([])

    with pytest.raises(BriefcaseCommandError, match="No publication channels"):
        publish_command(**options)


def test_multiple_channels_no_flag(publish_command, first_app):
    """Multiple channels without --channel raises an error."""
    publish_command.apps = {"first": first_app}

    with pytest.raises(BriefcaseCommandError, match="Specify a channel"):
        publish_command(channel=None)


def test_single_channel_auto_selects(publish_command, first_app):
    """A single available channel is used automatically."""
    publish_command._get_channels = lambda: {"only": _make_channel_class("only")}
    publish_command.apps = {"first": first_app}

    publish_command(channel=None)

    assert publish_command.actions == [
        ("verify-host",),
        ("verify-tools",),
        ("finalize-app-config", "first"),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("publish", "first", "only", {}),
    ]


def test_publish_triggers_package(publish_command, first_app_unpackaged):
    """If distribution artefact is missing, package is triggered first."""
    publish_command._get_channels = lambda: {"only": _make_channel_class("only")}
    publish_command.apps = {"first": first_app_unpackaged}
    # binary exists but distribution artefact does not

    publish_command(channel=None)

    assert ("package", "first") in [a[:2] for a in publish_command.actions]


def test_publish_with_update(publish_command, first_app):
    """An app update can be forced before publication."""
    publish_command._get_channels = lambda: {"only": _make_channel_class("only")}
    publish_command.apps = {"first": first_app}

    publish_command(update=True, channel=None)

    assert publish_command.actions == [
        ("verify-host",),
        ("verify-tools",),
        ("finalize-app-config", "first"),
        ("package", "first", {"update": True, "packaging_format": "Dummy"}),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("publish", "first", "only", {"package_state": "first"}),
    ]


def test_publish(publish_command, first_app, second_app):
    """If there are multiple apps, publish all of them."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Select a specific channel (required when multiple are available)
    options, _ = publish_command.parse_options(["-c", "s3"])

    # Run the publish command
    publish_command(**options)

    # The right sequence of things will be done
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Publish the first app to s3
        ("publish", "first", "s3", {}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Publish the second app to s3
        ("publish", "second", "s3", {"publish_state": "first"}),
    ]


def test_publish_alternative_channel(publish_command, first_app, second_app):
    """Apps can be published to an alternate channel."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = publish_command.parse_options(["-c", "alternative"])

    # Run the publish command
    publish_command(**options)

    # The right sequence of things will be done
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Publish the first app to the alternative channel
        ("publish", "first", "alternative", {}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Publish the second app to the alternative channel
        ("publish", "second", "alternative", {"publish_state": "first"}),
    ]


@pytest.mark.parametrize("app_flags", ["--app", "-a"])
def test_publish_app_single(publish_command, first_app, second_app, app_flags):
    """If the --app or -a flag is used, only the selected app is published."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options with the parameterized flag
    options, _ = publish_command.parse_options([app_flags, "first", "-c", "s3"])

    # Run the publish command
    publish_command(**options)

    # Only the selected app is published
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # Only the selected app config is finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified
        ("verify-app-tools", "first"),
        # Publish the selected app
        ("publish", "first", "s3", {}),
    ]


def test_publish_app_invalid(publish_command, first_app, second_app):
    """If an invalid app name is passed to --app, raise an error."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure the --app option with an invalid app
    options, _ = publish_command.parse_options(["--app", "invalid", "-c", "s3"])

    # Running the command should raise an error
    with pytest.raises(
        BriefcaseCommandError,
        match=r"App 'invalid' does not exist in this project.",
    ):
        publish_command(**options)


def test_publish_passes_packaging_format_to_package(
    publish_command,
    first_app_unpackaged,
):
    """packaging_format is forwarded to package_command when packaging is triggered."""
    publish_command._get_channels = lambda: {"only": _make_channel_class("only")}
    publish_command.apps = {"first": first_app_unpackaged}

    publish_command(channel=None)

    # Find the package action and verify packaging_format was passed
    package_actions = [a for a in publish_command.actions if a[0] == "package"]
    assert len(package_actions) == 1
    assert "packaging_format" in package_actions[0][2]
    assert package_actions[0][2]["packaging_format"] == "Dummy"


def test_non_existent(publish_command, first_app_config, second_app):
    """Publishing an app that hasn't been created cascades through packaging."""
    # Add two apps; use the "config only" version of the first app.
    publish_command.apps = {
        "first": first_app_config,
        "second": second_app,
    }

    options, _ = publish_command.parse_options(["-c", "s3"])

    publish_command(**options)

    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First app: no distribution artefact, so package is triggered
        ("package", "first", {"update": False, "packaging_format": "Dummy"}),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("publish", "first", "s3", {"package_state": "first"}),
        # Second app publishes normally (has distribution artefact)
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("publish", "second", "s3", {"publish_state": "first"}),
    ]


def test_unbuilt(publish_command, first_app_unbuilt, second_app):
    """Publishing an unbuilt app cascades through packaging."""
    publish_command.apps = {
        "first": first_app_unbuilt,
        "second": second_app,
    }

    options, _ = publish_command.parse_options(["-c", "s3"])

    publish_command(**options)

    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First app: no distribution artefact, so package is triggered first
        ("package", "first", {"update": False, "packaging_format": "Dummy"}),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("publish", "first", "s3", {"package_state": "first"}),
        # Second app publishes normally (has distribution artefact)
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("publish", "second", "s3", {"publish_state": "first"}),
    ]



def test_check_output_passes_through_to_subprocess(
    noop_context,
    mock_noop_subprocess_setup,
):
    """check_output passes all arguments to tools.subprocess.check_output."""
    mocks = mock_noop_subprocess_setup
    mock_output = "Python 3.10.0"
    mocks["subprocess"].check_output.return_value = mock_output

    args = ["python", "--version"]
    kwargs = {"encoding": "utf-8", "check": True}

    result = noop_context.check_output(args, **kwargs)

    mocks["subprocess"].check_output.assert_called_once_with(args, **kwargs)
    assert result == mock_output


def test_check_output_with_no_kwargs(noop_context, mock_noop_subprocess_setup):
    """check_output works with just args, no kwargs."""
    mocks = mock_noop_subprocess_setup
    mock_output = "output"
    mocks["subprocess"].check_output.return_value = mock_output

    args = ["echo", "test"]
    result = noop_context.check_output(args)

    mocks["subprocess"].check_output.assert_called_once_with(args)
    assert result == mock_output


def test_check_output_preserves_original_kwargs(
    noop_context,
    mock_noop_subprocess_setup,
):
    """check_output does not modify the original kwargs dict."""
    mocks = mock_noop_subprocess_setup
    mock_output = "output"
    mocks["subprocess"].check_output.return_value = mock_output

    original_kwargs = {"encoding": "utf-8", "env": {"CUSTOM": "value"}}
    kwargs_copy = original_kwargs.copy()

    result = noop_context.check_output(["test"], **original_kwargs)

    assert original_kwargs == kwargs_copy
    mocks["subprocess"].check_output.assert_called_once_with(
        ["test"], **original_kwargs
    )
    assert result == mock_output



import json
import platform
import subprocess
from unittest import mock

import pytest

from briefcase.console import LogLevel
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.windows.app import WindowsAppRunCommand


@pytest.fixture
def run_command(dummy_console, tmp_path):
    command = WindowsAppRunCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    command._stream_app_logs = mock.MagicMock()

    return command


def test_run_gui_app(run_command, first_app_config, tmp_path):
    """A Windows GUI app can be started."""
    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, passthrough=[])

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [tmp_path / "base_path/build/first-app/windows/app/src/First App.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )


def test_run_gui_app_with_passthrough(run_command, first_app_config, tmp_path):
    """A Windows GUI app can be started in debug mode with args."""
    run_command.console.verbosity = LogLevel.DEBUG

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )


def test_run_gui_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the GUI app, an exception is raised."""

    run_command.tools.subprocess.Popen.side_effect = OSError("Some error")

    with pytest.raises(OSError, match="Some error"):
        run_command.run_app(first_app_config, passthrough=[])

    # Popen was still invoked, though
    run_command.tools.subprocess.Popen.assert_called_with(
        [tmp_path / "base_path/build/first-app/windows/app/src/First App.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app(run_command, first_app_config, tmp_path):
    """A Windows GUI app can be started."""
    first_app_config.console_app = True

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, passthrough=[])

    # The process was started
    run_command.tools.subprocess.run.assert_called_with(
        [tmp_path / "base_path/build/first-app/windows/app/src/first-app.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        bufsize=1,
        stream_output=False,
    )

    # There is no streamer
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_with_passthrough(run_command, first_app_config, tmp_path):
    """A Windows console app can be started in debug mode with args."""
    run_command.console.verbosity = LogLevel.DEBUG

    first_app_config.console_app = True

    # Run the app with args
    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    run_command.tools.subprocess.run.assert_called_with(
        [
            tmp_path / "base_path/build/first-app/windows/app/src/first-app.exe",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        bufsize=1,
        stream_output=False,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # There is no streamer
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the console app, an exception is raised."""
    first_app_config.console_app = True

    run_command.tools.subprocess.run.side_effect = OSError("Some error")

    with pytest.raises(OSError, match="Some error"):
        run_command.run_app(first_app_config, passthrough=[])

    # Popen was still invoked, though
    run_command.tools.subprocess.run.assert_called_with(
        [tmp_path / "base_path/build/first-app/windows/app/src/first-app.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        bufsize=1,
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_app_test_mode(run_command, first_app_config, is_console_app, tmp_path):
    """A Windows app can be started in test mode."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app
    first_app_config.test_mode = True

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, passthrough=[])

    # The process was started
    exe_name = "first-app" if is_console_app else "First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [tmp_path / f"base_path/build/first-app/windows/app/src/{exe_name}.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_app_test_mode_with_passthrough(
    run_command,
    first_app_config,
    is_console_app,
    tmp_path,
):
    """A Windows app can be started in test mode with args."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app
    first_app_config.test_mode = True

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # The process was started
    exe_name = "first-app" if is_console_app else "First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path / f"base_path/build/first-app/windows/app/src/{exe_name}.exe",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )


def test_run_gui_app_debugger(run_command, first_app_config, tmp_path, dummy_debugger):
    """A Windows app can be started in debug mode."""
    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    first_app_config.debugger = dummy_debugger
    first_app_config.debugger_host = "somehost"
    first_app_config.debugger_port = 9999

    # Run the app
    run_command.run_app(first_app_config, passthrough=[])

    # The process was started
    run_command.tools.subprocess.Popen.assert_called_with(
        [tmp_path / "base_path/build/first-app/windows/app/src/First App.exe"],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={
            "BRIEFCASE_DEBUGGER": json.dumps(
                {
                    "debugger": "dummy",
                    "host": "somehost",
                    "port": 9999,
                    "host_os": platform.system(),
                    "app_path_mappings": {
                        "device_sys_path_regex": "app$",
                        "device_subfolders": ["first_app"],
                        "host_folders": [str(tmp_path / "base_path/src/first_app")],
                    },
                    "app_packages_path_mappings": None,
                }
            )
        },
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_on_icloud(dummy_command, first_app_templated, tmp_path):
    """If the path is on iCloud, the metadata is detected."""
    dummy_command.tools.subprocess.check_output.return_value = (
        "com.apple.fileprovider.fpfs#P: \n"
    )

    # Verify the location
    with pytest.raises(
        BriefcaseCommandError,
        match=r"app folder, move your project to location\nthat is not",
    ):
        dummy_command.verify_not_on_icloud(first_app_templated)

    # The subprocess call was what we expect
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xattr",
            "-p",
            "com.apple.fileprovider.fpfs#P",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        quiet=1,
    )

    # The bundle hasn't been cleaned up
    assert (tmp_path / "base_path/build/first-app/macos/app").exists()


def test_on_icloud_with_cleanup(dummy_command, first_app_templated, tmp_path):
    """If the path is on iCloud, the metadata is detected, and the bundle can be cleaned
    up."""
    dummy_command.tools.subprocess.check_output.return_value = (
        "com.apple.fileprovider.fpfs#P: \n"
    )

    # Verify the location
    with pytest.raises(
        BriefcaseCommandError,
        match=r"\n\nMove your project to a location that is not synchronized",
    ):
        dummy_command.verify_not_on_icloud(first_app_templated, cleanup=True)

    # The subprocess call was what we expect
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xattr",
            "-p",
            "com.apple.fileprovider.fpfs#P",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        quiet=1,
    )

    # The bundle has been cleaned up
    assert not (tmp_path / "base_path/build/first-app/macos/app").exists()


def test_not_on_icloud(dummy_command, first_app_templated, tmp_path):
    """If the path *not* is on iCloud, the metadata is detected."""
    dummy_command.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(returncode=1, cmd="xattr")
    )

    # The location can be verified
    dummy_command.verify_not_on_icloud(first_app_templated)

    # The subprocess call was what we expect
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xattr",
            "-p",
            "com.apple.fileprovider.fpfs#P",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        quiet=1,
    )

    # The bundle still exists
    assert (tmp_path / "base_path/build/first-app/macos/app").exists()


def test_app_doesnt_exist(dummy_command, first_app, tmp_path):
    """If the path hasn't been generated yet, verification succeeds."""
    dummy_command.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(returncode=1, cmd="xattr")
    )

    # The location can be verified
    dummy_command.verify_not_on_icloud(first_app)

    # The subprocess call was what we expect
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xattr",
            "-p",
            "com.apple.fileprovider.fpfs#P",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        quiet=1,
    )

    # The bundle doesn't exist
    assert not (tmp_path / "base_path/build/first-app/macos/app").exists()



from unittest import mock

from ...utils import PartialMatchString


def test_question_sequence(new_command):
    """Questions are asked, a context is constructed."""

    # Prime answers for all the questions.
    new_command.console.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
    ]

    context = new_command.build_app_context(
        project_overrides={},
    )

    assert context == {
        "app_name": "myapplication",
        "author": "Grace Hopper",
        "author_email": "grace@navy.mil",
        "bundle": "org.beeware",
        "class_name": "MyApplication",
        "description": "Cool stuff",
        "formal_name": "My Application",
        "license": "GPL-2.0",
        "module_name": "myapplication",
        "source_dir": "src/myapplication",
        "test_source_dir": "tests",
        "project_name": "My Project",
        "url": "https://navy.mil/myapplication",
    }


def test_question_sequence_with_overrides(new_command):
    """Overrides can be used to set the answers for questions."""

    # Prime answers for none of the questions.
    new_command.console.values = []

    context = new_command.build_app_context(
        project_overrides={
            "formal_name": "My Override App",
            "app_name": "myoverrideapp",
            "bundle": "net.example",
            "project_name": "My Override Project",
            "description": "My override description",
            "author": "override, author",
            "author_email": "author@override.tld",
            "url": "https://override.example.com",
            "license": "MIT",
        },
    )

    assert context == {
        "app_name": "myoverrideapp",
        "author": "override, author",
        "author_email": "author@override.tld",
        "bundle": "net.example",
        "class_name": "MyOverrideApp",
        "description": "My override description",
        "formal_name": "My Override App",
        "license": "MIT",
        "module_name": "myoverrideapp",
        "source_dir": "src/myoverrideapp",
        "test_source_dir": "tests",
        "project_name": "My Override Project",
        "url": "https://override.example.com",
    }


def test_question_sequence_with_bad_license_override(new_command):
    """A bad override for license uses user input instead."""

    # Prime answers for all the questions.
    new_command.console.values = [
        "4",  # license
    ]

    context = new_command.build_app_context(
        project_overrides={
            "formal_name": "My Override App",
            "app_name": "myoverrideapp",
            "bundle": "net.example",
            "project_name": "My Override Project",
            "description": "My override description",
            "author": "override, author",
            "author_email": "author@override.tld",
            "url": "https://override.example.com",
            "license": "BAD i don't exist license",
        },
    )

    assert context == {
        "app_name": "myoverrideapp",
        "author": "override, author",
        "author_email": "author@override.tld",
        "bundle": "net.example",
        "class_name": "MyOverrideApp",
        "description": "My override description",
        "formal_name": "My Override App",
        "license": "GPL-2.0",
        "module_name": "myoverrideapp",
        "source_dir": "src/myoverrideapp",
        "test_source_dir": "tests",
        "project_name": "My Override Project",
        "url": "https://override.example.com",
    }


def test_question_sequence_with_no_user_input(new_command):
    """If no user input is provided, all user inputs are taken as default."""

    new_command.console.input_enabled = False

    context = new_command.build_app_context(project_overrides={})

    assert context == {
        "app_name": "helloworld",
        "author": "Jane Developer",
        "author_email": "jane@example.com",
        "bundle": "com.example",
        "class_name": "HelloWorld",
        "description": "My first application",
        "formal_name": "Hello World",
        "license": "BSD-3-Clause",
        "module_name": "helloworld",
        "source_dir": "src/helloworld",
        "test_source_dir": "tests",
        "project_name": "Hello World",
        "url": "https://example.com/helloworld",
    }


def test_author_and_email_use_git_config_as_fallback(new_command):
    """If no user input is provided, git config values 'git.user' and 'git.email' are
    used if available."""
    new_command.tools.git = object()
    new_command.get_git_config_value = mock.MagicMock()
    new_command.get_git_config_value.side_effect = ["Some Author", "my@email.com"]

    new_command.console.input_enabled = False

    context = new_command.build_app_context(project_overrides={})

    assert context["author"] == "Some Author"
    assert context["author_email"] == "my@email.com"
    assert new_command.get_git_config_value.call_args_list == [
        mock.call("user", "name"),
        mock.call("user", "email"),
    ]


def test_git_config_is_mentioned_as_source(new_command, monkeypatch):
    """If git config is used as default value, this shall be mentioned to the user."""
    new_command.tools.git = object()
    new_command.get_git_config_value = mock.MagicMock()
    new_command.get_git_config_value.side_effect = ["Some Author", "my@email.com"]

    new_command.console.input_enabled = False

    mock_text_question = mock.MagicMock()
    mock_text_question.side_effect = lambda *args, **kwargs: kwargs["default"]
    monkeypatch.setattr(new_command.console, "text_question", mock_text_question)

    new_command.build_app_context(project_overrides={})

    assert (
        mock.call(
            intro=PartialMatchString("Based on your git configuration"),
            description="Author",
            default="Some Author",
            override_value=None,
        )
        in mock_text_question.call_args_list
    )

    assert (
        mock.call(
            intro=PartialMatchString("Based on your git configuration"),
            description="Author's Email",
            default="my@email.com",
            override_value=None,
            validator=new_command.validate_email,
        )
        in mock_text_question.call_args_list
    )



from __future__ import annotations

import subprocess
from collections.abc import Collection
from pathlib import Path

from packaging.version import Version

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.base import Tool, ToolCache
from briefcase.integrations.subprocess import SubprocessArgT


class Flatpak(Tool):
    name = "flatpak"
    full_name = "Flatpak"
    supported_host_os: Collection[str] = {"Linux"}

    DEFAULT_REPO_ALIAS = "flathub"
    DEFAULT_REPO_URL = "https://flathub.org/repo/flathub.flatpakrepo"

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> Flatpak:
        """Verify that the Flatpak toolchain is available.

        :param tools: ToolCache of available tools
        :returns: A wrapper for the Flatpak tools.
        """
        # short circuit since already verified and available
        if hasattr(tools, "flatpak"):
            return tools.flatpak

        flatpak = Flatpak(tools=tools)
        try:
            output = tools.subprocess.check_output(["flatpak", "--version"]).strip("\n")
            parts = output.split(" ")
            try:
                if parts[0] == "Flatpak":
                    version = parts[1].split(".")
                    if int(version[0]) < 1:
                        raise BriefcaseCommandError(
                            "Briefcase requires Flatpak 1.0 or later."
                        )
                else:
                    raise ValueError(f"Unexpected tool name {parts[0]}") from None
            except (ValueError, IndexError):
                tools.console.warning(
                    """\
*************************************************************************
** WARNING: Unable to determine the version of Flatpak                 **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you
    experience problems, this is almost certainly the cause of those
    problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

      flatpak --version

    from the command prompt.

*************************************************************************
"""
                )

        except OSError as e:
            raise BriefcaseCommandError(
                """\
Briefcase requires the Flatpak toolchain, but it does not appear to be installed.

Instructions for installing the Flatpak toolchain can be found at:

    https://flatpak.org/setup/

You must install both flatpak and flatpak-builder.
"""
            ) from e
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Unable to invoke flatpak.") from e

        try:
            output = tools.subprocess.check_output(
                ["flatpak-builder", "--version"]
            ).strip("\n")

            # flatpak-builder 1.3 changed the output of --version
            # from "flatpak-builder 1.2.X" to "flatpak-build-1.3.X".
            # Converge on the new-style format.
            parts = output.replace(" ", "-").rsplit("-", 1)
            try:
                if parts[0] == "flatpak-builder":
                    version = parts[1].split(".")
                    if int(version[0]) < 1:
                        raise BriefcaseCommandError(
                            "Briefcase requires flatpak-builder 1.0 or later."
                        )
                else:
                    raise ValueError(f"Unexpected tool name {parts[0]}")
            except (ValueError, IndexError):
                tools.console.warning(
                    """\
*************************************************************************
** WARNING: Unable to determine the version of flatpak-builder         **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you
    experience problems, this is almost certainly the cause of those
    problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

      flatpak-builder --version

    from the command prompt.

*************************************************************************
"""
                )

        except OSError as e:
            raise BriefcaseCommandError(
                """\
Briefcase requires the full Flatpak development toolchain, but flatpak-builder
does not appear to be installed.

Instructions for installing the Flatpak toolchain can be found at:

    https://flatpak.org/setup/

You must install both flatpak and flatpak-builder.
"""
            ) from e
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Unable to invoke flatpak-builder.") from e

        tools.flatpak = flatpak
        return flatpak

    def verify_repo(self, repo_alias: str, url: str):
        """Verify that the Flatpak repository has been registered.

        :param repo_alias: The alias to use when registering the repo.
        :param url: The URL of the Flatpak repo.
        """
        try:
            self.tools.subprocess.run(
                [
                    "flatpak",
                    "remote-add",
                    "--user",
                    "--if-not-exists",
                    repo_alias,
                    url,
                ]
                + (["--verbose"] if self.tools.console.is_deep_debug else []),
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to add Flatpak repo {url} with alias {repo_alias}."
            ) from e

    def verify_runtime(
        self,
        repo_alias: str,
        runtime: str,
        runtime_version: str,
        sdk: str,
    ):
        """Verify that a specific Flatpak runtime and SDK are available.

        :param repo_alias: The alias of the repo where the runtime and SDK are stored.
        :param runtime: The identifier of the Flatpak runtime
        :param runtime_version: The version of the Flatpak runtime
        :param sdk: The Flatpak SDK
        """
        try:
            self.tools.subprocess.run(
                [
                    "flatpak",
                    "install",
                    "--assumeyes",
                    "--user",
                    repo_alias,
                    f"{runtime}/{self.tools.host_arch}/{runtime_version}",
                    f"{sdk}/{self.tools.host_arch}/{runtime_version}",
                ]
                + (["--verbose"] if self.tools.console.is_deep_debug else []),
                check=True,
                # flatpak install uses many animations that cannot be disabled
                stream_output=False,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to install "
                f"Flatpak runtime {runtime}/{self.tools.host_arch}/{runtime_version} "
                f"and SDK {sdk}/{self.tools.host_arch}/{runtime_version} "
                f"from repo {repo_alias}."
            ) from e

    def build(self, bundle_identifier: str, app_name: str, path: Path):
        """Build a Flatpak manifest.

        On success, the app is installed into the user's local Flatpak install, and a
        shell script is created that can be used to start the app. The shell file isn't
        really needed to start the app, but it serves as a marker for a successful build
        that Briefcase can use.

        :param bundle_identifier: The bundle identifier for the app being built.
        :param app_name: The app name.
        :param path: The path to the folder containing the app's Flatpak manifest file.
        """
        try:
            self.tools.subprocess.run(
                [
                    "flatpak-builder",
                    "--force-clean",
                    # Archive into a local repository
                    "--repo",
                    "repo",
                    # Install the app into the user space
                    "--install",
                    "--user",
                    "build",
                    "manifest.yml",
                ]
                + (["--verbose"] if self.tools.console.is_deep_debug else []),
                check=True,
                cwd=path,
            )

            # Create a marker file to indicate a build has completed.
            # For bonus points, the marker file also is executable
            # and is an alias for the command that would actually start
            # the flatpak.
            bin_path = path / bundle_identifier
            with bin_path.open("w", encoding="utf-8") as f:
                f.write(
                    f"""\
#!/bin/sh
# echo To run this flatpak, run:
flatpak run {bundle_identifier}
"""
                )
                self.tools.os.chmod(bin_path, 0o755)
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(f"Error while building app {app_name}.") from e

    def run(
        self,
        bundle_identifier: str,
        args: list[SubprocessArgT] | None = None,
        stream_output: bool = True,
        **kwargs,
    ) -> subprocess.Popen[str]:
        """Run a Flatpak in a way that allows for log streaming.

        :param bundle_identifier: The bundle identifier for the app being built.
        :param args: (Optional) The list of arguments to pass to the app
        :param stream_output: Should output be streamed?
        :returns: A Popen object for the running app; or ``None`` if the app isn't
            streaming
        """
        flatpak_run_cmd = ["flatpak", "run", bundle_identifier]
        flatpak_run_cmd.extend([] if args is None else args)

        if self.tools.console.is_deep_debug:
            # Must come before bundle identifier;
            # otherwise, it's passed as an arg to app
            flatpak_run_cmd.insert(2, "--verbose")

        if stream_output:
            return self.tools.subprocess.Popen(
                flatpak_run_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
                **kwargs,
            )
        else:
            return self.tools.subprocess.run(
                flatpak_run_cmd,
                bufsize=1,
                stream_output=False,
                **kwargs,
            )

    def bundle(
        self,
        repo_url: str,
        bundle_identifier: str,
        app_name: str,
        version: Version,
        build_path: Path,
        output_path: Path,
    ):
        """Bundle a Flatpak for distribution.

        Generates a standalone .flatpak file that can be installed into another user's
        Flatpak repository.

        :param repo_url: The URL of the repository that contains the runtime used by the
            app.
        :param bundle_identifier: The bundle identifier for the app being built.
        :param app_name: The app name.
        :param version: The version of the app being built.
        :param build_path: The path where the flatpak was built. This path will contain
            the repo where the built flatpak was exported.
        :param output_path: The path of the output file to create as an export.
        """
        try:
            self.tools.subprocess.run(
                [
                    "flatpak",
                    "build-bundle",
                    # Set the repo where the runtime can be found
                    "--runtime-repo",
                    repo_url,
                    # Sign the export
                    # "--gpg-sign", "..."
                    "repo",
                    output_path,
                    bundle_identifier,
                    str(version),
                ]
                + (["--verbose"] if self.tools.console.is_deep_debug else []),
                check=True,
                cwd=build_path,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to build a Flatpak bundle for app {app_name}."
            ) from e



from __future__ import annotations

import re
import subprocess
import uuid
from collections.abc import Collection
from pathlib import Path, PurePath
from zipfile import ZIP_DEFLATED, ZipFile

from briefcase.commands import CreateCommand, PackageCommand, RunCommand
from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.windows_sdk import WindowsSDK
from briefcase.integrations.wix import WiX

DEFAULT_OUTPUT_FORMAT = "app"


def txt_to_rtf(txt):
    """A very simple TXT to RTF converter.

    The entire document is rendered in Courier. Any blank line is interpreted as a
    paragraph marker; any line starting with a * is rendered as a bullet. Everything
    else is rendered verbatim in the RTF document.

    :param text: The original text.
    :returns: The text in RTF format.
    """
    rtf = ["{\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}"]
    for line in txt.split("\n"):
        if line.lstrip().startswith("*"):
            rtf.append(f"\\bullet{line[line.index('*') + 1 :]} ")
        elif line:
            # Add a space at the end to ensure multi-line paragraphs
            # have a word break. Strip whitespace to ensure that
            # indented bullet paragraphs don't have extra space.
            rtf.append(line.strip() + " ")
        else:
            # A blank line is a paragraph+line break.
            rtf.append("\\par\\line")
    rtf.append("}")

    return "\n".join(rtf)


class WindowsMixin:
    platform = "windows"
    supported_host_os: Collection[str] = {"Windows"}
    supported_host_os_reason = "Windows applications can only be built on Windows."
    platform_target_version = "0.3.24"

    def bundle_package_executable_path(self, app):
        if app.console_app:
            return f"{app.app_name}.exe"
        else:
            return f"{app.formal_name}.exe"

    def bundle_package_path(self, app):
        return self.bundle_path(app) / self.packaging_root

    def binary_path(self, app):
        return self.package_path(app) / self.package_executable_path(app)

    def distribution_path(self, app):
        suffix = "zip" if app.packaging_format == "zip" else "msi"
        return self.dist_path / f"{app.formal_name}-{app.version}.{suffix}"

    def verify_host(self):
        super().verify_host()
        # The stub app only supports x86-64 right now, and our VisualStudio and WiX code
        # is the same (#1887). However, we can package an external x86-64 app on any
        # build machine.
        if self.tools.host_arch != "AMD64":
            if all(app.external_package_path for app in self.apps.values()):
                if not self.is_clone:
                    self.console.warning(f"""
*************************************************************************
** WARNING: Possible architecture mismatch                             **
*************************************************************************

The build machine is {self.tools.host_arch}, but Briefcase on Windows currently only
supports x86-64 installers.

You are responsible for ensuring that the content of external_package_path
is compatible with x86-64.

*************************************************************************
""")
            else:
                raise UnsupportedHostError(
                    "Windows applications cannot be built on an "
                    f"{self.tools.host_arch} machine."
                )

        # 64bit Python is required to ensure 64bit wheels are installed/created
        # for the app
        if self.tools.is_32bit_python:
            raise UnsupportedHostError("""\
Windows applications cannot be built using a 32bit version of Python.

Install a 64bit version of Python and run Briefcase again.
""")


class WindowsCreateCommand(CreateCommand):
    def support_package_filename(self, support_revision):
        return f"python-{self.python_version_tag}.{support_revision}-embed-amd64.zip"

    def support_package_url(self, support_revision):
        micro = re.match(r"\d+", str(support_revision)).group(0)
        return (
            "https://www.python.org/ftp/python/"
            f"{self.python_version_tag}.{micro}/"
            f"{self.support_package_filename(support_revision)}"
        )

    def extras_path(self, app: AppConfig) -> Path:
        """Obtain the path for extra installer content.

        Extra installer content is content that needs to be inserted into the installer,
        in addition to material in the packaging path. It is primarily used to add
        installer scripts when packaging an externally managed app.

        :param app: The config object for the app
        :return: The full path where install scripts should be installed.
        """
        try:
            extras_path = self.path_index(app, "extras_path")
        except KeyError:
            # For backwards compatibility - if the template doesn't define an extras
            # path, default to `extras`.
            extras_path = "extras"

        return self.bundle_path(app) / extras_path

    def output_format_template_context(self, app: AppConfig):
        """Additional template context required by the output format.

        :param app: The config object for the app
        """
        # WiX requires a 3-element, integer-only version number. If a version
        # triple isn't explicitly provided, generate one by stripping any
        # non-numeric portions from the version number.
        # If there are less than 3 numeric parts, 0s will be appended.
        try:
            version_triple = app.version_triple
        except AttributeError:
            version_triple = ".".join(
                ([str(v) for v in app.version.release] + ["0", "0"])[:3]
            )

        # The application needs a unique GUID.
        # This is used to track the application, even if the application
        # name changes. We can generate a default GUID using the bundle
        # and the formal name; but you'll need to manually set this value
        # if you ever change those two keys.
        try:
            guid = app.guid
        except AttributeError:
            # Create a DNS domain by reversing the bundle identifier
            domain = ".".join(app.bundle_identifier.split(".")[::-1])
            guid = uuid.uuid5(uuid.NAMESPACE_DNS, domain)
            self.console.info(f"Assigning {app.app_name} an application GUID of {guid}")

        try:
            install_scope = "perMachine" if app.system_installer else "perUser"
        except AttributeError:
            # system_installer not defined in config; default to asking the user
            install_scope = "perUserOrMachine"

        return {
            "version_triple": version_triple,
            "guid": str(guid),
            "install_scope": install_scope,
            "package_path": str(self.package_path(app)),
            "binary_path": self.package_executable_path(app),
        }

    def _cleanup_app_support_package(self, support_path):
        # On Windows, the support path is co-mingled with app content.
        # This means updating the support package is imperfect.
        # Warn the user that there could be problems.
        self.console.warning("""
*************************************************************************
** WARNING: Support package update may be imperfect                    **
*************************************************************************

    Support packages in Windows apps are overlaid with app content,
    so it isn't possible to remove all old support files before
    installing new ones.

    Briefcase will unpack the new support package without cleaning up
    existing support package content. This *should* work; however,
    ensure a reproducible release artefacts, it is advisable to
    perform a clean app build before release.

*************************************************************************
""")

    def install_license(self, app: AppConfig):
        """Install the license for the project as RTF content.

        Currently assumes PEP621 format for `license`:
        * If `license.file` is an RTF file, it is used verbatim
        * If `license.file` is any other file, it is converted to RTF
          using a simple text->RTF conversion.
        * If `license.text` is provided, that text is converted to
          RTF; with a warning for the case where `license.text` is
          a one-line license name/description.

        If no `license` field is defined, or it points at a file that
        doesn't exist an error is raised.

        When PEP639 support is added, we will need to adapt this method.

        :param app: The config object for the app
        """
        installed_license = self.bundle_path(app) / "LICENSE.rtf"

        if license_file := app.license.get("file"):
            license_file = self.base_path / license_file
            if license_file.is_file():
                if license_file.suffix == ".rtf":
                    self.tools.shutil.copy(license_file, installed_license)
                    license_text = None
                else:
                    license_text = license_file.read_text(encoding="utf-8")
                    installed_license.write_text(
                        txt_to_rtf(license_text), encoding="utf-8"
                    )
            else:
                raise BriefcaseCommandError(
                    "Your `pyproject.toml` specifies a license file of "
                    f"{str(license_file.relative_to(self.base_path))!r}.\n"
                    "However, this file does not exist."
                    "\n\n"
                    "Ensure you have correctly spelled the filename in your "
                    "`license.file` setting."
                )
        elif license_text := app.license.get("text"):
            if len(license_text.splitlines()) <= 1:
                self.console.warning("""
Your app specifies a license using `license.text`, but the value doesn't appear
to be a full license. Briefcase will generate a `LICENSE.rtf` file for your
project; you should ensure that the contents of this file is adequate.
""")
            installed_license.write_text(
                txt_to_rtf(license_text),
                encoding="utf-8",
            )
        else:
            raise BriefcaseCommandError("""\
Your project does not contain a `license` definition.

Create a file named `LICENSE` in the same directory as your `pyproject.toml`
with your app's licensing terms, and set `license.file = 'LICENSE'` in your
app's configuration.
""")

    def install_app_resources(self, app: AppConfig):
        """Install Windows-specific app resources.

        This includes any post-install or pre-uninstall scripts, plus converting the
        LICENSE file into an RTF file.

        :param app: The config object for the app
        """
        super().install_app_resources(app)

        installer_path = getattr(app, "installer_path", "_installer")
        install_scripts_path = self.extras_path(app) / installer_path

        # Ensure the extras path exists, so that the path used by WiX exists
        self.extras_path(app).mkdir(exist_ok=True, parents=True)

        # Install the post-install script
        if post_install := getattr(app, "post_install_script", None):
            post_install_script = self.base_path / post_install
            if post_install_script.suffix != ".bat":
                raise BriefcaseCommandError(
                    "Windows post-install scripts must be .bat files."
                )
            elif not post_install_script.is_file():
                raise BriefcaseCommandError(
                    f"Couldn't find post-install script {post_install}."
                )

            with self.console.wait_bar("Installing post-install script..."):
                install_scripts_path.mkdir(exist_ok=True, parents=True)
                self.tools.shutil.copyfile(
                    post_install_script,
                    install_scripts_path / "post_install.bat",
                )

        # Install the pre-uninstall script
        if pre_uninstall := getattr(app, "pre_uninstall_script", None):
            pre_uninstall_script = self.base_path / pre_uninstall
            if pre_uninstall_script.suffix != ".bat":
                raise BriefcaseCommandError(
                    "Windows pre-uninstall scripts must be .bat files."
                )
            elif not pre_uninstall_script.is_file():
                raise BriefcaseCommandError(
                    f"Couldn't find pre-uninstall script {pre_uninstall}."
                )

            with self.console.wait_bar("Installing pre-uninstall script..."):
                install_scripts_path.mkdir(exist_ok=True, parents=True)
                self.tools.shutil.copyfile(
                    pre_uninstall_script,
                    install_scripts_path / "pre_uninstall.bat",
                )

        # Install the license.
        with self.console.wait_bar("Installing license..."):
            self.install_license(app)


class WindowsRunCommand(RunCommand):
    supports_debugger = True

    def run_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        """
        # Set up the log stream
        kwargs = self._prepare_app_kwargs(app=app)

        # Console apps must operate in non-streaming mode so that console input can
        # be handled correctly. However, if we're in test mode, we *must* stream so
        # that we can see the test exit sentinel
        if app.console_app and not app.test_mode:
            self.console.info("=" * 75)
            self.tools.subprocess.run(
                [self.binary_path(app), *passthrough],
                cwd=self.tools.home_path,
                encoding="UTF-8",
                bufsize=1,
                stream_output=False,
                **kwargs,
            )
        else:
            # Start the app in a way that lets us stream the logs
            app_popen = self.tools.subprocess.Popen(
                [self.binary_path(app), *passthrough],
                cwd=self.tools.home_path,
                encoding="UTF-8",
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
                **kwargs,
            )

            # Start streaming logs for the app.
            self._stream_app_logs(
                app,
                popen=app_popen,
                clean_output=False,
            )


class WindowsPackageCommand(PackageCommand):
    ADHOC_SIGN_HELP = (
        "Perform no signing on the app. "
        "Your app will be reported as coming from an unverified publisher."
    )

    IDENTITY_HELP = "The 40-digit hex checksum of the code signing identity to use"

    @property
    def packaging_formats(self):
        return ["msi", "zip"]

    @property
    def default_packaging_format(self):
        return "msi"

    def verify_tools(self):
        super().verify_tools()
        WiX.verify(tools=self.tools)
        if self._windows_sdk_needed:
            WindowsSDK.verify(tools=self.tools)

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "--file-digest",
            help="File digest algorithm to use for code signing; defaults to sha256",
            default="sha256",
            required=False,
        )
        parser.add_argument(
            "--use-local-machine-stores",
            help=(
                "Specifies the code signing certificate is stored in the "
                "Local Machine's stores instead of the Current User's"
            ),
            action="store_true",
            dest="use_local_machine",
            required=False,
        )
        parser.add_argument(
            "--cert-store",
            help=(
                "The internal Windows name for the certificate store containing the "
                "certificate for code signing; defaults to 'My' for the Personal store"
            ),
            default="My",
            required=False,
        )
        parser.add_argument(
            "--timestamp-url",
            help=(
                "URL for the Timestamp Authority server to timestamp the code signing; "
                "defaults to timestamp.digicert.com"
            ),
            default="http://timestamp.digicert.com",
            required=False,
        )
        parser.add_argument(
            "--timestamp-digest",
            help=(
                "Digest algorithm to request the Timestamp Authority server uses "
                "for the timestamp for code signing; defaults to sha256"
            ),
            default="sha256",
            required=False,
        )

    def parse_options(self, extra):
        """Require the Windows SDK tool if an `identity` is specified for signing."""
        options, overrides = super().parse_options(extra=extra)
        self._windows_sdk_needed = options["identity"] is not None
        return options, overrides

    def sign_file(
        self,
        app: AppConfig,
        filepath: Path,
        identity: str,
        file_digest: str,
        use_local_machine: bool,
        cert_store: str,
        timestamp_url: str,
        timestamp_digest: str,
    ):
        """Sign a file."""

        if not re.fullmatch(r"^[0-9a-f]{40}$", identity, flags=re.IGNORECASE):
            raise BriefcaseCommandError(
                f"Codesigning identify {identity!r} must be a "
                "certificate SHA-1 thumbprint."
            )

        sign_command = [
            self.tools.windows_sdk.signtool_exe,
            "sign",
            "-s",
            cert_store,
            "-sha1",
            identity,
            "-fd",
            file_digest,
            "-d",
            app.description,
            "-du",
            app.url,
            "-tr",
            timestamp_url,
            "-td",
            timestamp_digest,
        ]

        if use_local_machine:
            # Look for cert in Local Machine instead of Current User
            sign_command.append("-sm")

        # Filepath to sign must come last
        sign_command.append(filepath)

        try:
            self.tools.subprocess.run(sign_command, check=True)
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(f"Unable to sign {filepath}") from e

    def package_app(
        self,
        app: AppConfig,
        identity: str | None = None,
        adhoc_sign: bool = False,
        file_digest: str | None = None,
        use_local_machine: bool = False,
        cert_store: str | None = None,
        timestamp_url: str | None = None,
        timestamp_digest: str | None = None,
        **kwargs,
    ):
        """Package an application.

        Code signing parameters are ignored if ``sign_app=False``. If a code signing
        identity is not provided, then ``sign_app=False`` will be enforced.

        :param app: The application to package
        :param identity: SHA-1 thumbprint of the certificate to use for code signing.
        :param adhoc_sign: Should the application be signed? Default: ``True``
        :param file_digest: File hashing algorithm for code signing.
        :param use_local_machine: True to use cert stores for the Local Machine instead
            of the Current User; default to False for Current User.
        :param cert_store: Certificate store within Current User or Local Machine to
            search for the certificate within.
        :param timestamp_url: Timestamp authority server to use in code signing.
        :param timestamp_digest: Hashing algorithm to request from the timestamp server.
        """

        if adhoc_sign:
            sign_app = False
        elif identity:
            sign_app = True
        else:
            sign_app = False
            self.console.warning("""
*************************************************************************
** WARNING: No signing identity provided                               **
*************************************************************************

    Briefcase will not sign the app. To provide a signing identity,
    use the `--identity` option; or, to explicitly disable signing,
    use `--adhoc-sign`.

*************************************************************************
""")

        if sign_app:
            self.console.info("Signing App...", prefix=app.app_name)
            sign_options = {
                "identity": identity,
                "file_digest": file_digest,
                "use_local_machine": use_local_machine,
                "cert_store": cert_store,
                "timestamp_url": timestamp_url,
                "timestamp_digest": timestamp_digest,
            }
            self.sign_file(app=app, filepath=self.binary_path(app), **sign_options)

        if app.packaging_format == "zip":
            self._package_zip(app)
        else:
            self._package_msi(app)

            if sign_app:
                self.console.info("Signing MSI...", prefix=app.app_name)
                self.sign_file(
                    app=app,
                    filepath=self.distribution_path(app),
                    **sign_options,
                )

    def _package_msi(self, app):
        """Build the msi installer."""
        try:
            with self.console.wait_bar("Building MSI..."):
                self.tools.subprocess.run(
                    [
                        self.tools.wix.wix_exe,
                        "build",
                        "-ext",
                        self.tools.wix.ext_path("UI"),
                        "-arch",
                        "x64",  # Default is x86, regardless of the build machine.
                        f"{app.app_name}.wxs",
                        "-loc",
                        "unicode.wxl",
                        "-pdbtype",
                        "none",
                        "-o",
                        self.distribution_path(app),
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(f"Unable to package app {app.app_name}.") from e

    def _package_zip(self, app):
        """Package the app as simple zip file."""

        self.console.info("Building zip file...", prefix=app.app_name)
        with self.console.wait_bar("Packing..."):
            source = self.package_path(app)
            zip_root = f"{app.formal_name}-{app.version}"

            with ZipFile(self.distribution_path(app), "w", ZIP_DEFLATED) as archive:
                for file_path in source.glob("**/*"):
                    archive.write(
                        file_path, zip_root / PurePath(file_path).relative_to(source)
                    )



from unittest.mock import MagicMock


def test_package_zip(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
):
    """A macOS App can be packaged as a zip."""
    # Mock the creation of the ditto archive
    package_command.ditto_archive = MagicMock()

    # Select zip packaging
    first_app_with_binaries.packaging_format = "zip"

    # Select a code signing identity
    package_command.select_identity.return_value = sekrit_identity

    # Package the app in zip (not DMG) format
    package_command.package_app(first_app_with_binaries)

    # A request has been made to sign the app
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=sekrit_identity,
    )

    # A request has been made to notarize the app
    package_command.notarize.assert_called_once_with(
        first_app_with_binaries,
        identity=sekrit_identity,
    )

    # No dmg was built.
    assert package_command.dmgbuild.build_dmg.call_count == 0

    # If the DMG doesn't exist, it can't be signed either.
    # This ignores the calls that would have been made transitively
    # by calling sign_app()
    assert package_command.sign_file.call_count == 0

    # The notarization process will create the final artefact;
    # since we're mocking notarize, we rely on the notarization tests
    # to verify that the final distribution artefact was created.


def test_zip_no_notarization(
    package_command,
    sekrit_identity,
    first_app_with_binaries,
    tmp_path,
):
    """A macOS App can be packaged as a zip, without notarization."""
    # Mock the creation of the ditto archive
    package_command.ditto_archive = MagicMock()

    # Select zip packaging
    first_app_with_binaries.packaging_format = "zip"

    # Select a code signing identity
    package_command.select_identity.return_value = sekrit_identity

    # Package the app in zip (not DMG) format, disabling notarization
    package_command.package_app(
        first_app_with_binaries,
        notarize_app=False,
    )

    # A request has been made to sign the app
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=sekrit_identity,
    )

    # No request has been made to notarize the app
    package_command.notarize.assert_not_called()

    # No dmg was built.
    assert package_command.dmgbuild.build_dmg.call_count == 0

    # If the DMG doesn't exist, it can't be signed either.
    # This ignores the calls that would have been made transitively
    # by calling sign_app()
    assert package_command.sign_file.call_count == 0

    # Since we're not notarizing, the packaged archive was created.
    package_command.ditto_archive.assert_called_once_with(
        tmp_path / "base_path/build/first-app/macos/app/First App.app",
        tmp_path / "base_path/dist/First App-0.0.1.app.zip",
    )


def test_external_app(
    package_command,
    external_first_app,
    sekrit_identity,
):
    """An external macOS App can be packaged as a zip."""
    # Mock the creation of the ditto archive
    package_command.ditto_archive = MagicMock()

    # Select zip packaging
    external_first_app.packaging_format = "zip"

    # Select a code signing identity
    package_command.select_identity.return_value = sekrit_identity

    # Package the app in zip (not DMG) format
    package_command.package_app(external_first_app)

    # A request has been made to sign the app
    package_command.sign_app.assert_called_once_with(
        app=external_first_app,
        identity=sekrit_identity,
    )

    # A request has been made to notarize the app
    package_command.notarize.assert_called_once_with(
        external_first_app,
        identity=sekrit_identity,
    )

    # No dmg was built.
    assert package_command.dmgbuild.build_dmg.call_count == 0

    # If the DMG doesn't exist, it can't be signed either.
    # This ignores the calls that would have been made transitively
    # by calling sign_app()
    assert package_command.sign_file.call_count == 0

    # The notarization process will create the final artefact;
    # since we're mocking notarize, we rely on the notarization tests
    # to verify that the final distribution artefact was created.



import pytest

from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return WindowsAppCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
